// ----------------------------------
// Filename      : Link.java
// Created       : 08/10/1999
// Last modified : 22/06/2000
// Target        : Java VM
// Author        : Sven Maerivoet
//
// All rights reserved.
// Universitaire Instelling Antwerpen
// ----------------------------------

package mitrasim.network;

import java.util.*;
import mitrasim.engine3d.*;
import mitrasim.network.*;
import mitrasim.util.*;

/** @author Sven Maerivoet<BR>
    All rights reserved.<BR>
    Universitaire Instelling Antwerpen
*/
public class Link
{
  // the different link-types 
  public static final short kTunnel           = 1;
  public static final short kHighway          = 2;
  public static final short kSeparatedOnRamp  = 3;
  public static final short kMergedOnRamp     = 4;
  public static final short kSeparatedOffRamp = 5;
  public static final short kMergedOffRamp    = 6;
  public static final short kNarrowing        = 7;
  public static final short kWidening         = 8;

  // the different connection-parts
  public static final short kConnectedToMainLink = 1;
  public static final short kConnectedToOnRamp   = 2;

  // the different locations for narrowings, widenings and ramps
  public static final short kLeft  = 1;
  public static final short kRight = 2;

  // the different lanechange-permissions
  public static final short kAllowed    = 1;
  public static final short kNotAllowed = 2;

  public int     fID;
  public int     fNrOfNodes;
  public Node[]  fNodes;
  public int     fNrOfLanes;
  public int[]   fNrOfLanesAtNode;
  public int[]   fLengths;
  public float[] fGrades;
  public int     fSpeedLimit;
  public boolean fIsTerminator;
  public int     fMainLinkConnectedToLink;
  public int     fMainLinkConnectedToPart;
  public short   fType;
  public short   fLocation;
  public int     fOffRampConnectedToLink;
  public int     fOffRampConnectedToPart;
  public int     fNrOfLanesOnRamp;
  public int     fNewNrOfLanes;
  public int     fNrOfLaneChangePermissions;
  public int[]   fLaneChangePermissions;
  public String  fName;
  public Link    fMainLinksPreviousLink; // who is connected to my mainlink ?
  public Link    fOnRampsPreviousLink; // who is connected to my onramp ?
  public Link    fMainLinksNextLink; // to whom am I connected with my mainlink ?
  public Link    fOffRampsNextLink; // to whom am I connected with my offramp ?
  public boolean fConnectionToMyMainLinkIsAlreadyDefined;
  public boolean fConnectionToMyOnRampIsAlreadyDefined;
  public boolean fTopologicallySorted;
  public Vector  fVehicleQueue;
  public float[] fSegmentAverageSpeed;
  public float[] fSegmentDensity;
  public float   fRampLength;

  public Link()
  {
    fID = 0;
    fNrOfNodes = 0;
    fNodes = null;
    fNrOfLanes = 1;
    fNrOfLanesAtNode = null;
    fLengths = null;
    fGrades = null;
    fSpeedLimit = 120;
    fIsTerminator = false;
    fMainLinkConnectedToLink = 0;
    fMainLinkConnectedToPart = 0;
    fType = kTunnel;
    fLocation = kLeft;
    fOffRampConnectedToLink = 0;
    fOffRampConnectedToPart = 0;
    fNrOfLanesOnRamp = 1;
    fNewNrOfLanes = 1;
    fNrOfLaneChangePermissions = 2;
    fLaneChangePermissions = null;
    fName = "";
    fMainLinksPreviousLink = null;
    fOnRampsPreviousLink = null;
    fMainLinksNextLink = null;
    fOffRampsNextLink = null;
    fConnectionToMyMainLinkIsAlreadyDefined = false;
    fConnectionToMyOnRampIsAlreadyDefined = false;
    fTopologicallySorted = false;
    fVehicleQueue = null;
    fSegmentAverageSpeed = null;    
    fSegmentDensity = null;
    fRampLength = 0.0f;
  }

  public String translate(Messages messages)
  {
    if (fType == kTunnel) {
      return messages.lookup("textTunnelLink");
    }
    else if (fType == kHighway) {
      return messages.lookup("textHighwayLink");
    }
    else if (fType == kSeparatedOnRamp) {
      return messages.lookup("textSeparatedOnRampLink");
    }
    else if (fType == kMergedOnRamp) {
      return messages.lookup("textMergedOnRampLink");
    }
    else if (fType == kSeparatedOffRamp) {
      return messages.lookup("textSeparatedOffRampLink");
    }
    else if (fType == kMergedOffRamp) {
      return messages.lookup("textMergedOffRampLink");
    }
    else if (fType == kNarrowing) {
      return messages.lookup("textNarrowingLink");
    }
    else if (fType == kWidening) {
      return messages.lookup("textWideningLink");
    }
    return ""; // the compiler complains without this final return-statement
  }

  public void show()
  {
    System.out.println("Link #" + fID + " : type=" + " " + fType + "[" + fName + "]");
  }

  public void calcGeometry()
  {
    final int SplineInterpolation = 1024;

    float Cx[] = new float[4];
    float Cy[] = new float[4];
    float Cz[] = new float[4];

    Point3D[] P = new Point3D[4];
    for (int i = 0; i < 4; ++i) {
      P[i] = new Point3D();
    }

    SplineIntegrator SplineLength = new SplineIntegrator();

    for (int i = 0; i < (fNrOfNodes - 1); ++i) {

      if (i == 0) {
        // handle the first spline-segment
        if (fMainLinksPreviousLink != null) {
          convertNode(fMainLinksPreviousLink.fNodes[fMainLinksPreviousLink.fNrOfNodes - 2],P[0]);
        }
        else if (fOnRampsPreviousLink != null) {
          convertNode(fOnRampsPreviousLink.fNodes[fOnRampsPreviousLink.fNrOfNodes - 2],P[0]);
        }
        else {
          // duplicate first node
          convertNode(fNodes[0],P[0]);
        }
        convertNode(fNodes[0],P[1]);
        convertNode(fNodes[1],P[2]);
        if (fNrOfNodes > 2) {
          convertNode(fNodes[2],P[3]);
        }
        else {
          if (fMainLinksNextLink != null) {
            convertNode(fMainLinksNextLink.fNodes[1],P[3]);
          }
          else if (fOffRampsNextLink != null) {
            convertNode(fOffRampsNextLink.fNodes[1],P[3]);
          }
          else {
            // duplicate last node
            convertNode(fNodes[1],P[3]);
          }
        }
      }
      else if (i < (fNrOfNodes - 2)) {
        // handle the intermediate spline-segments
        if (fNrOfNodes >= 4) {
          convertNode(fNodes[i - 1],P[0]);
          convertNode(fNodes[i],P[1]);
          convertNode(fNodes[i + 1],P[2]);
          convertNode(fNodes[i + 2],P[3]);
        }
      }
      else {
        // handle the last spline-segment
        if (fNrOfNodes > 2) {
          convertNode(fNodes[fNrOfNodes - 3],P[0]);
        }
        else {
          if (fMainLinksPreviousLink != null) {
            convertNode(fMainLinksPreviousLink.fNodes[fMainLinksPreviousLink.fNrOfNodes - 2],P[0]);
          }
          else if (fOnRampsPreviousLink != null) {
            convertNode(fOnRampsPreviousLink.fNodes[fOnRampsPreviousLink.fNrOfNodes - 2],P[0]);
          }
          else {
            // duplicate first node
            convertNode(fNodes[0],P[0]);
          }
        }
        convertNode(fNodes[fNrOfNodes - 2],P[1]);
        convertNode(fNodes[fNrOfNodes - 1],P[2]);
        if (fMainLinksNextLink != null) {
          convertNode(fMainLinksNextLink.fNodes[1],P[3]);
        }
        else if (fOffRampsNextLink != null) {
          convertNode(fOffRampsNextLink.fNodes[1],P[3]);
        }
        else {
          // duplicate last node
          convertNode(fNodes[fNrOfNodes - 1],P[3]);
        }
      }

      // calculate the product of the Catmull-Rom base-matrix and
      // the 4 controlpoints (that will be interpolated)

      // the x-coordinate
      Cx[0] = (-1 * P[0].fX) +
              ( 3 * P[1].fX) +
              (-3 * P[2].fX) +
              ( 1 * P[3].fX);
      Cx[1] = ( 2 * P[0].fX) +
              (-5 * P[1].fX) +
              ( 4 * P[2].fX) +
              (-1 * P[3].fX);
      Cx[2] = (-1 * P[0].fX) +
              ( 1 * P[2].fX);
      Cx[3] = ( 2 * P[1].fX);

      // the y-coordinate
      Cy[0] = (-1 * P[0].fY) +
              ( 3 * P[1].fY) +
              (-3 * P[2].fY) +
              ( 1 * P[3].fY);
      Cy[1] = ( 2 * P[0].fY) +
              (-5 * P[1].fY) +
              ( 4 * P[2].fY) +
              (-1 * P[3].fY);
      Cy[2] = (-1 * P[0].fY) +
              ( 1 * P[2].fY);
      Cy[3] = ( 2 * P[1].fY);

      // the z-coordinate
      Cz[0] = (-1 * P[0].fZ) +
              ( 3 * P[1].fZ) +
              (-3 * P[2].fZ) +
              ( 1 * P[3].fZ);
      Cz[1] = ( 2 * P[0].fZ) +
              (-5 * P[1].fZ) +
              ( 4 * P[2].fZ) +
              (-1 * P[3].fZ);
      Cz[2] = (-1 * P[0].fZ) +
              ( 1 * P[2].fZ);
      Cz[3] = ( 2 * P[1].fZ);

      for (int j = 0; j < 4; ++j) {
        Cx[j] = (Cx[j] / 2.0f);
        Cy[j] = (Cy[j] / 2.0f);
        Cz[j] = (Cz[j] / 2.0f);
      }

      // calculate the length of each segment in the link
      SplineLength.a = (9 * Cx[0] * Cx[0]) +
                       (9 * Cy[0] * Cy[0]) +
                       (9 * Cz[0] * Cz[0]);
      SplineLength.b = (12 * Cx[0] * Cx[1]) +
                       (12 * Cy[0] * Cy[1]) +
                       (12 * Cz[0] * Cz[1]);
      SplineLength.c = ((6 * Cx[0] * Cx[2]) + (4 * Cx[1] * Cx[1])) +
                       ((6 * Cy[0] * Cy[2]) + (4 * Cy[1] * Cy[1])) +
                       ((6 * Cz[0] * Cz[2]) + (4 * Cz[1] * Cz[1]));
      SplineLength.d = (4 * Cx[1] * Cx[2]) +
                       (4 * Cy[1] * Cy[2]) +
                       (4 * Cz[1] * Cz[2]);
      SplineLength.e = (Cx[2] * Cx[2]) +
                       (Cy[2] * Cy[2]) +
                       (Cz[2] * Cz[2]);

      fLengths[i] = (int) SplineLength.integrate(0.0f,1.0f,SplineInterpolation);

      // calculate the slope of each segment in the link
      // (e.g.: slope = 15% means that over a distance of 100 m the
      // link inclines for 15 m)
      fGrades[i] = (100.0f / fLengths[i]) * (fNodes[i + 1].fZPos - fNodes[i].fZPos);
    }

    fNrOfLanesAtNode = new int[fNrOfNodes];
    // calculate the number of lanes at each node
    for (int NodeNr = 0; NodeNr < fNrOfNodes; ++NodeNr) {

      fNrOfLanesAtNode[NodeNr] = fNrOfLanes;

      // it's ok for tunnels and highways, now check the other possibilities
      if (fType == kSeparatedOnRamp) {
        fNrOfLanesAtNode[NodeNr] = fNrOfLanes + fNrOfLanesOnRamp;
      }
      else if (fType == kMergedOnRamp) {
        if (NodeNr < (fNrOfNodes - 1)) {
          fNrOfLanesAtNode[NodeNr] = fNrOfLanes + fNrOfLanesOnRamp;
        }
        else {
          fNrOfLanesAtNode[NodeNr] = fNrOfLanes;
        }
      }
      else if (fType == kSeparatedOffRamp) {
        if (NodeNr == 0) {
          fNrOfLanesAtNode[NodeNr] = fNrOfLanes;
        }
        else {
          fNrOfLanesAtNode[NodeNr] = fNrOfLanes + fNrOfLanesOnRamp;
        }
      }
      else if (fType == kMergedOffRamp) {
        fNrOfLanesAtNode[NodeNr] = fNrOfLanes;
      }
      else if (fType == kNarrowing) {
        if (NodeNr < (fNrOfNodes - 1)) {
          fNrOfLanesAtNode[NodeNr] = fNrOfLanes;
        }
        else {
          fNrOfLanesAtNode[NodeNr] = fNewNrOfLanes;
        }
      }
      else if (fType == kWidening) {
        if (NodeNr == 0) {
          fNrOfLanesAtNode[NodeNr] = fNrOfLanes;
        }
        else {
          fNrOfLanesAtNode[NodeNr] = fNewNrOfLanes;
        }
      }
    }

    fSegmentAverageSpeed = new float[fNrOfNodes - 1];
    for (int i = 0; i < (fNrOfNodes - 1); ++i) {
      fSegmentAverageSpeed[i] = 0.0f;
    }

    fSegmentDensity = new float[fNrOfNodes - 1];
    for (int i = 0; i < (fNrOfNodes - 1); ++i) {
      fSegmentDensity[i] = 0.0f;
    }

    fRampLength = 0.0f;
    if ((fType == kSeparatedOnRamp) || (fType == kMergedOnRamp)) {
      for (int i = 0; i < (fNrOfNodes - 2); ++i) {
        fRampLength += fLengths[i];
      }
    }
    else if ((fType == kSeparatedOffRamp) || (fType == kMergedOffRamp)) {
      for (int i = 1; i < (fNrOfNodes - 1); ++i) {
        fRampLength += fLengths[i];
      }
    }
  }

  private void convertNode(Node node, Point3D point)
  {
    point.set(node.fXPos,node.fYPos,node.fZPos);
  }


  private double sqr(double n)
  {
    return (n * n);
  }

  public void cloneTo(Link duplicate)
  {
    duplicate.fID = fID;
    duplicate.fNrOfNodes = fNrOfNodes;
    if (fNodes != null) {
      duplicate.fNodes = new Node[fNrOfNodes];
      for (int i = 0; i < fNrOfNodes; ++i) {
        duplicate.fNodes[i] = fNodes[i];
      }
    }
    duplicate.fNrOfLanes = fNrOfLanes;
    if (fNrOfLanesAtNode != null) {
      duplicate.fNrOfLanesAtNode = new int[fNrOfLanesAtNode.length];
      for (int i = 0; i < fNrOfLanesAtNode.length; ++i) {
        duplicate.fNrOfLanesAtNode[i] = fNrOfLanesAtNode[i];
      }
    }
    if (fLengths != null) {
      duplicate.fLengths = new int[fNrOfNodes - 1];
      for (int i = 0; i < (fNrOfNodes - 1); ++i) {
        duplicate.fLengths[i] = fLengths[i];
      }
    }
    if (fGrades != null) {
      duplicate.fGrades = new float[fNrOfNodes - 1];
      for (int i = 0; i < (fNrOfNodes - 1); ++i) {
        duplicate.fGrades[i] = fGrades[i];
      }
    }
    duplicate.fSpeedLimit = fSpeedLimit;
    duplicate.fIsTerminator = fIsTerminator;
    duplicate.fMainLinkConnectedToLink = fMainLinkConnectedToLink;
    duplicate.fMainLinkConnectedToPart = fMainLinkConnectedToPart;
    duplicate.fType = fType;
    duplicate.fLocation = fLocation;
    duplicate.fOffRampConnectedToLink = fOffRampConnectedToLink;
    duplicate.fOffRampConnectedToPart = fOffRampConnectedToPart;
    duplicate.fNrOfLanesOnRamp = fNrOfLanesOnRamp;
    duplicate.fNewNrOfLanes = fNewNrOfLanes;
    duplicate.fNrOfLaneChangePermissions = fNrOfLaneChangePermissions;
    if (fLaneChangePermissions != null) {
      duplicate.fLaneChangePermissions = new int[fNrOfLaneChangePermissions];
      for (int i = 0; i < fNrOfLaneChangePermissions; ++i) {
        duplicate.fLaneChangePermissions[i] = fLaneChangePermissions[i];
      }
    }
    duplicate.fName = new String(fName);
    duplicate.fMainLinksPreviousLink = fMainLinksPreviousLink;
    duplicate.fOnRampsPreviousLink = fOnRampsPreviousLink;
    duplicate.fMainLinksNextLink = fMainLinksNextLink;
    duplicate.fOffRampsNextLink = fOffRampsNextLink;
    duplicate.fConnectionToMyMainLinkIsAlreadyDefined = fConnectionToMyMainLinkIsAlreadyDefined;
    duplicate.fConnectionToMyOnRampIsAlreadyDefined = fConnectionToMyOnRampIsAlreadyDefined;
    duplicate.fTopologicallySorted = fTopologicallySorted;
    duplicate.fVehicleQueue = fVehicleQueue; // warning : reference copied !
    if (fSegmentAverageSpeed != null) {
      duplicate.fSegmentAverageSpeed = new float[fSegmentAverageSpeed.length];
      for (int i = 0; i < fSegmentAverageSpeed.length; ++i) {
        duplicate.fSegmentAverageSpeed[i] = fSegmentAverageSpeed[i];
      }
    }
    if (fSegmentDensity != null) {
      duplicate.fSegmentDensity = new float[fSegmentDensity.length];
      for (int i = 0; i < fSegmentDensity.length; ++i) {
        duplicate.fSegmentDensity[i] = fSegmentDensity[i];
      }
    }
    duplicate.fRampLength = fRampLength;
  }

  public boolean nodeCompliant()
  {
    return (fNrOfNodes >= 2);
  }

  public void createDefaultLaneChangePermissions()
  {
    if ((fType == kTunnel) ||
        (fType == kHighway) ||
        (fType == kNarrowing)) {
      fNrOfLaneChangePermissions = 2 * (fNrOfLanes - 1);
    }
    else if ((fType == kSeparatedOnRamp) ||
             (fType == kMergedOnRamp) ||
             (fType == kSeparatedOffRamp) ||
             (fType == kMergedOffRamp)) {
      fNrOfLaneChangePermissions = 2 * (fNrOfLanesOnRamp + fNrOfLanes - 1);
    }
    else if (fType == kWidening) {
      fNrOfLaneChangePermissions = 2 * (fNewNrOfLanes - 1);
    }

    fLaneChangePermissions = new int[fNrOfLaneChangePermissions];

    // allow all lanechanges
    for (int i = 0; i < fNrOfLaneChangePermissions; ++i) {
      fLaneChangePermissions[i] = kAllowed;
    }    

    // tunnels, highways, separated onramps and widenings are already handled
    // so we now handle the other types
    if (fType == kMergedOnRamp) {
      if (fLocation == kLeft) {
        int Offset = 2 * (fNrOfLanes - 1);
        int Limit = fNrOfLanesOnRamp;
        for (int i = 0; i < Limit; ++i) {
          fLaneChangePermissions[Offset + (2 * i)] = kNotAllowed;
        }
      }
      else if (fLocation == kRight) {
        int Limit = fNrOfLanesOnRamp;
        for (int i = 0; i < Limit; ++i) {
          fLaneChangePermissions[(2 * i) + 1] = kNotAllowed;
        }
      }
    }
    else if (fType == kSeparatedOffRamp) {
      if (fLocation == kLeft) {
        int Index = (2 * (fNrOfLanes - 1)) + 1;
        fLaneChangePermissions[Index] = kNotAllowed;
      }
      else if (fLocation == kRight) {
        int Index = 2 * (fNrOfLanesOnRamp - 1);
        fLaneChangePermissions[Index] = kNotAllowed;
      }
    }
    else if (fType == kMergedOffRamp) {
      if (fLocation == kLeft) {
        int Index = (2 * ((fNrOfLanes - fNrOfLanesOnRamp) - 1)) + 1;
        fLaneChangePermissions[Index] = kNotAllowed;
      }
      else if (fLocation == kRight) {
        int Index = 2 * (fNrOfLanesOnRamp - 1);
        fLaneChangePermissions[Index] = kNotAllowed;
      }
    }
    else if (fType == kNarrowing) {
      if (fLocation == kLeft) {
        int Offset = 2 * (fNewNrOfLanes - 1);
        int Limit = fNrOfLanes - fNewNrOfLanes;
        for (int i = 0; i < Limit; ++i) {
          fLaneChangePermissions[Offset + (2 * i)] = kNotAllowed;
        }
      }
      else if (fLocation == kRight) {
        int Limit = fNrOfLanes - fNewNrOfLanes;
        for (int i = 0; i < Limit; ++i) {
          fLaneChangePermissions[(2 * i) + 1] = kNotAllowed;
        }
      }
    }
  }

  /** @author Sven Maerivoet<BR>
      All rights reserved.<BR>
      Universitaire Instelling Antwerpen
  */
  private class SplineIntegrator extends Integrator
  {
    float a, b, c, d, e;

    public SplineIntegrator()
    {
      super();
    }

    public float integrandum(float x)
    {
      return (float) Math.sqrt(e + x * (d + x * (c + x * (b + x * a))));
    }
  }
}
