// ------------------------------------------
// Filename      : RemoteControlProtocol.java
// Created       : 07/03/2000
// Last modified : 20/06/2000
// Target        : Java VM
// Author        : Sven Maerivoet
//
// All rights reserved.
// Universitaire Instelling Antwerpen
// ------------------------------------------

package mitrasim.server;

import java.io.*;
import java.net.*;
import java.text.*;
import java.util.*;
import mitrasim.network.*;
import mitrasim.resources.*;
import mitrasim.simulator.*;
import mitrasim.util.*;

/** @author Sven Maerivoet<BR>
    All rights reserved.<BR>
    Universitaire Instelling Antwerpen
*/
public class RemoteControlProtocol extends Protocol
{
  // the different messages
  public static final int kDisconnect             = 0;
  public static final int kShowConnectionLog      = 1;
  public static final int kUploadScenario         = 2;
  public static final int kSetCycleHoldTime       = 3;
  public static final int kAcquireState           = 4;
  public static final int kStartContinueSimulator = 5;
  public static final int kStopSimulator          = 6;
  public static final int kResetSimulator         = 7;
  public static final int kShutdownServer         = 8;

  // internal datastructures
  private Resources fResources;
  private Simulator fSimulator;

  public RemoteControlProtocol(InetAddress IP, DataInputStream in, DataOutputStream out)
  {
    super(IP,in,out);
    fResources = new Resources();
    fSimulator = new Simulator();
  }

  synchronized public void communicateWithClient()
  {
    try {
      // retrieve the client's request
      int Request = fIn.readInt();

      if (Request == kDisconnect) {
        showMessage(fMessages.lookup("messageDisconnectReceived"));

        fOut.writeBoolean(true);
        fOut.flush();

        // get the date and time of disconnect
        Date CurrentDate = new Date();
        String Timestamp = DateFormat.getDateTimeInstance(DateFormat.SHORT,DateFormat.MEDIUM).format(CurrentDate);
        String ReportString = fMessages.lookup("messageLogDisconnect",
                                               Timestamp,
                                               String.valueOf(fIP),
                                               getProtocolAbbreviation());

        // log the disconnect
        fResources.fServerLog.addLine(ReportString);

        fIsOperational = false;
      }
      else if (Request == kShowConnectionLog) {
        showMessage(fMessages.lookup("messageShowConnectionLogReceived"));
        showMessage(fMessages.lookup("messageSendingConnectionLog"));
        fResources.fServerLog.writeToStream(fOut);
      }
      else if (Request == kUploadScenario) {
        showMessage(fMessages.lookup("messageUploadScenarioReceived"));

        fSimulator.prepareForScenarioUpload();
        showMessageNoNewLine(fMessages.lookup("messageReceivingNetwork"));
        fSimulator.readNetworkFromStream(fIn);
        System.out.println();

        // check the network
        boolean NetworkOk = fSimulator.checkNetwork();
        fOut.writeBoolean(NetworkOk);
        fOut.flush();

        if (!NetworkOk) {
          showMessage(fMessages.lookup("errorScenarioUploadFailed"));
        }
        else {
          fSimulator.calcNetworkGeometry();
          showMessage(fMessages.lookup("messageScenarioUploadCompleted"));
        }

        // set the timestamp of the network
        fSimulator.setTimeStamp();

        // topologically sort the network
        showMessageNoNewLine(fMessages.lookup("messageTopologicallySortingNetwork"));
        fSimulator.topologicallySortNetwork();
        System.out.println(" " + fMessages.lookup("messageOk"));

        // receive the simulation parameters
        fResources.fsimStartTime = fIn.readInt();
        fResources.fsimPercentageMiddleLaneDrivers = fIn.readInt();
        fResources.fsimReactionTimeMean = fIn.readFloat();
        fResources.fsimReactionTimeStdDev = fIn.readFloat();
        fResources.fsimReactionTimeMin = fIn.readFloat();
        fResources.fsimReactionTimeMax = fIn.readFloat();
        fResources.fsimInitialVehicleSpeed = fIn.readFloat();
        fResources.fsimCycleHoldTime = fIn.readInt();
        fResources.fsimMinSpaceAtEntrance = fIn.readFloat();
        fResources.fsimAggressionAndExcessCoupled = fIn.readBoolean();
        fResources.fsimGravitationalConstant = fIn.readFloat();
        fResources.fsimCarFollowingMinSpaceHeadway = fIn.readFloat();
        fResources.fsimCarFollowingMinSpaceHeadwayExcess = fIn.readFloat();
        fResources.fsimCarFollowingMinTimeHeadway = fIn.readFloat();
        fResources.fsimCarFollowingMinTimeHeadwayExcess = fIn.readFloat();
        fResources.fsimCarFollowingMaxTimeHeadway = fIn.readFloat();
        fResources.fsimCarFollowingMaxTimeHeadwayExcess = fIn.readFloat();
        fResources.fsimCarFollowingAlphaMin = fIn.readFloat();
        fResources.fsimCarFollowingBetaMin = fIn.readFloat();
        fResources.fsimCarFollowingGammaMin = fIn.readFloat();
        fResources.fsimCarFollowingAlphaPlus = fIn.readFloat();
        fResources.fsimCarFollowingBetaPlus = fIn.readFloat();
        fResources.fsimCarFollowingGammaPlus = fIn.readFloat();
        fResources.fsimCriticalRampDistance = fIn.readFloat();
        fResources.fsimPowerOfMLCProbability = fIn.readInt();
        fResources.fsimLaneChangingMinLeadGap = fIn.readFloat();
        fResources.fsimLaneChangingMinLeadGapExcess = fIn.readFloat();
        fResources.fsimLaneChangingMinLagGap = fIn.readFloat();
        fResources.fsimLaneChangingMinLagGapExcess = fIn.readFloat();
        fResources.fsimAllowTrucksOnLanes = fIn.readInt();
        fResources.fsimMinTimeLeftDuringOvertaking = fIn.readInt();

        showMessage(fMessages.lookup("messageReceivingRoutingTable"));

        // receive the routing table
        fResources.fsimRoutingTable = new Vector();
        int NrOfRoutingTableEntries = fIn.readInt();
        for (int i = 0; i < NrOfRoutingTableEntries; ++i) {
          RoutingTableEntry routingTableEntry = new RoutingTableEntry();

          int NodeNr = fIn.readInt();
          routingTableEntry.fOriginNode = (Node) fResources.fNetwork.fNodes.get(new Integer(NodeNr));
          routingTableEntry.fFlowRate = fIn.readInt();
          routingTableEntry.fDistributionFactor = fIn.readInt();
          routingTableEntry.fFleetMix = fIn.readInt();
          routingTableEntry.fNrOfRoutes = fIn.readInt();
          routingTableEntry.fPercentages = new int[routingTableEntry.fNrOfRoutes];
          for (int j = 0; j < routingTableEntry.fNrOfRoutes; ++j) {
            routingTableEntry.fPercentages[j] = fIn.readInt();
          }
          routingTableEntry.fRoutes = new Node[routingTableEntry.fNrOfRoutes][];
          for (int j = 0; j < routingTableEntry.fNrOfRoutes; ++j) {
            int NrOfNodes = fIn.readInt();
            routingTableEntry.fRoutes[j] = new Node[NrOfNodes];
            for (int k = 0; k < NrOfNodes; ++k) {
              NodeNr = fIn.readInt();
              routingTableEntry.fRoutes[j][k] = (Node) fResources.fNetwork.fNodes.get(new Integer(NodeNr));
            }
          }

          // find the link associated with this routing table entry
          boolean Found = false;
          Enumeration LinkEnum = fResources.fNetwork.fLinks.elements();
          while (LinkEnum.hasMoreElements() && !Found) {
            Link link = (Link) LinkEnum.nextElement();

            if (link.fNodes[0].fID == routingTableEntry.fOriginNode.fID) {
              routingTableEntry.fNetworkEntranceLink = link;
              Found = true;
            }
          }

          // add the routing table entry to the routing table
          fResources.fsimRoutingTable.addElement(routingTableEntry);
        }

        // create the vehicle generators
        fSimulator.createVehicleGenerators();

        showMessage(fMessages.lookup("messageReceivingLoopDetectors"));

        // receive the loopdetectors
        // reserve room for 20 loopdetectors and a load factor of 80%
        fResources.fsimLoopDetectors = new Hashtable(20,0.8f);

        int NrOfLoopDetectors = fIn.readInt();
        for (int i = 0; i < NrOfLoopDetectors; ++i) {

          String Name = fIn.readUTF();
          int NodeID = fIn.readInt();
          int TimeForMeasurement = fIn.readInt();

          Node Location = (Node) fResources.fNetwork.fNodes.get(new Integer(NodeID));
          LoopDetector loopDetector = new LoopDetector(Name,Location,TimeForMeasurement);
          fResources.fsimLoopDetectors.put(new Integer(NodeID),loopDetector);
        }

        showMessage(fMessages.lookup("messageReceivingViewSpecifications"));

        // receive the view-specifications
        // reserve room for 20 view-specifications and a load factor of 80%
        fResources.fvisViewSpecifications = new Hashtable(20,0.8f);

        int NrOfViewSpecifications = fIn.readInt();
        for (int i = 0; i < NrOfViewSpecifications; ++i) {
          ViewSpecification viewSpecification = new ViewSpecification();

          viewSpecification.fName = fIn.readUTF();
          viewSpecification.fPerspectiveMode = fIn.readBoolean();
          viewSpecification.fViewPointX = fIn.readInt();
          viewSpecification.fViewPointY = fIn.readInt();
          viewSpecification.fViewPointZ = fIn.readInt();
          viewSpecification.fEyePointRho = fIn.readInt();
          viewSpecification.fEyePointPhi = fIn.readInt();
          viewSpecification.fEyePointTheta = fIn.readInt();
          viewSpecification.fBoundingBoxMinX = fIn.readInt();
          viewSpecification.fBoundingBoxMinY = fIn.readInt();
          viewSpecification.fBoundingBoxMaxX = fIn.readInt();
          viewSpecification.fBoundingBoxMaxY = fIn.readInt();
          viewSpecification.fShowEnvironment = fIn.readBoolean();
          fResources.fvisViewSpecifications.put(new String(viewSpecification.fName),viewSpecification);
        }

        fSimulator.resetSimulator();
      }
      else if (Request == kSetCycleHoldTime) {
        showMessage(fMessages.lookup("messageSetCycleHoldTimeReceived"));

        fResources.fsimCycleHoldTime = fIn.readInt();
      }
      else if (Request == kAcquireState) {
        showMessage(fMessages.lookup("messageAcquireSimulatorStateReceived"));

        fOut.writeBoolean(fSimulator.isSimulatorReady());
        fOut.writeBoolean(fSimulator.isSimulatorRunning());
        fOut.flush();
      }
      else if (Request == kStartContinueSimulator) {
        showMessage(fMessages.lookup("messageStartContinueSimulatorReceived"));

        fOut.writeBoolean(fSimulator.startContinueSimulator());
        fOut.flush();
      }
      else if (Request == kStopSimulator) {
        showMessage(fMessages.lookup("messageStopSimulatorReceived"));

        fOut.writeBoolean(fSimulator.stopSimulator());
        fOut.flush();
      }
      else if (Request == kResetSimulator) {
        showMessage(fMessages.lookup("messageResetSimulatorReceived"));

        fOut.writeBoolean(fSimulator.resetSimulator());
        fOut.flush();
      }
      else if (Request == kShutdownServer) {
        showMessage(fMessages.lookup("messageShutdownServerReceived"));

        fOut.writeBoolean(true);
        fOut.flush();

        // terminate the server
        System.exit(0);
      }
      else {
        showMessage(fMessages.lookup("errorInvalidRequestReceived"));

        // indicate request-failure
        fOut.writeBoolean(false);
        fOut.flush();
      }
    }
    catch (IOException exc) {
      showMessage(fMessages.lookup("errorCommunicationFailure"));

      // get the date and time of disconnect
      Date CurrentDate = new Date();
      String Timestamp = DateFormat.getDateTimeInstance(DateFormat.SHORT,DateFormat.MEDIUM).format(CurrentDate);
      String ReportString = fMessages.lookup("messageLogDisconnect",
                                             Timestamp,
                                             String.valueOf(fIP),
                                             getProtocolAbbreviation());

      // log the disconnect
      fResources.fServerLog.addLine(ReportString);

      fIsOperational = false;
    }
  }

  public int getProtocol()
  {
    return kRemoteControlProtocol;
  }

  public String getProtocolAbbreviation()
  {
    return fMessages.lookup("textRemoteControlProtocolAbbreviation");;
  }

  public String translate(Messages messages)
  {
    return messages.lookup("textRemoteControlProtocol");
  }
}
