// ---------------------------------------
// Filename      : VisualizerProtocol.java
// Created       : 22/04/2000
// Last modified : 21/06/2000
// Target        : Java VM
// Author        : Sven Maerivoet
//
// All rights reserved.
// Universitaire Instelling Antwerpen
// ---------------------------------------

package mitrasim.server;

import java.io.*;
import java.net.*;
import java.text.*;
import java.util.*;
import mitrasim.network.*;
import mitrasim.resources.*;
import mitrasim.simulator.*;
import mitrasim.util.*;

/** @author Sven Maerivoet<BR>
    All rights reserved.<BR>
    Universitaire Instelling Antwerpen
*/
public class VisualizerProtocol extends Protocol
{
  // the different messages
  public static final int kDownloadNetwork           = 0;
  public static final int kSetActiveLinks            = 1;
  public static final int kRequestData               = 2;
  public static final int kDisconnect                = 3;

  // internal datastructures
  private Resources fResources;
  private Simulator fSimulator;
  private Vector    fActiveLinks;

  public VisualizerProtocol(InetAddress IP, DataInputStream in, DataOutputStream out)
  {
    super(IP,in,out);
    fResources = new Resources();
    fSimulator = new Simulator();
    fActiveLinks = null;
  }

  synchronized public void communicateWithClient()
  {
    try {
      // retrieve the client's request
      int Request = fIn.readInt();

      if (Request == kDownloadNetwork) {
        // indicate readiness to the visualizer
        if (fSimulator.isSimulatorReady()) {
          fOut.writeBoolean(true);
          fOut.flush();
        }
        else {
          fOut.writeBoolean(false);
          fOut.flush();
          return;
        }

        showMessage(fMessages.lookup("messageSendingNetworkToVisualizer"));
        fSimulator.writeNetworkToStream(fOut);

        showMessage(fMessages.lookup("messageSendingTimeStampToVisualizer"));
        fSimulator.writeTimeStampToStream(fOut);

        // upload the view-specifications
        showMessage(fMessages.lookup("messageSendingViewSpecificationsToVisualizer"));

        fOut.writeInt(fResources.fvisViewSpecifications.size());
        Enumeration ViewSpecificationEnum = fResources.fvisViewSpecifications.elements();
        while (ViewSpecificationEnum.hasMoreElements()) {
          ViewSpecification viewSpecification = (ViewSpecification) ViewSpecificationEnum.nextElement();

          fOut.writeUTF(viewSpecification.fName);
          fOut.writeBoolean(viewSpecification.fPerspectiveMode);
          fOut.writeInt(viewSpecification.fViewPointX);
          fOut.writeInt(viewSpecification.fViewPointY);
          fOut.writeInt(viewSpecification.fViewPointZ);
          fOut.writeInt(viewSpecification.fEyePointRho);
          fOut.writeInt(viewSpecification.fEyePointPhi);
          fOut.writeInt(viewSpecification.fEyePointTheta);
          fOut.writeInt(viewSpecification.fBoundingBoxMinX);
          fOut.writeInt(viewSpecification.fBoundingBoxMinY);
          fOut.writeInt(viewSpecification.fBoundingBoxMaxX);
          fOut.writeInt(viewSpecification.fBoundingBoxMaxY);
          fOut.writeBoolean(viewSpecification.fShowEnvironment);
        }
        fOut.flush();
      }
      else if (Request == kSetActiveLinks) {
        showMessage(fMessages.lookup("messageSetActiveLinksReceived"));

        fActiveLinks = new Vector();

        int NrOfActiveLinks = fIn.readInt();
        for (int i = 0; i < NrOfActiveLinks; ++i) {
          int LinkID = fIn.readInt();

          // search and add the specified link to the list of active links
          Link link = (Link) fResources.fNetwork.fLinks.get(new Integer(LinkID));
          fActiveLinks.addElement(link);
        }
      }
      else if (Request == kRequestData) {
        // send the timestamp of the network
        fSimulator.writeTimeStampToStream(fOut);

        // send the simulator's status and current time
        fOut.writeBoolean(fSimulator.isSimulatorRunning());
        fOut.writeInt(fSimulator.getMasterClock());

        // send all the vehicles on the active links
        Enumeration ActiveLinkEnum = fActiveLinks.elements();
        while (ActiveLinkEnum.hasMoreElements()) {
          Link link = (Link) ActiveLinkEnum.nextElement();

          synchronized (this) {

            // copy the vehicles to a temporary space for synchronization with the simulator
            Vector TempVehicleQueue = new Vector();

            Enumeration VehicleEnum = link.fVehicleQueue.elements();
            try {
              while (VehicleEnum.hasMoreElements()) {
                Vehicle vehicle = (Vehicle) VehicleEnum.nextElement();

                Vehicle TempVehicle = new Vehicle();
                vehicle.cloneTo(TempVehicle);

                TempVehicleQueue.addElement(TempVehicle);
              }
            }
            catch (NoSuchElementException exc) {
            }

            fOut.writeInt(TempVehicleQueue.size());

            // process all the vehicles
            VehicleEnum = TempVehicleQueue.elements();
            while (VehicleEnum.hasMoreElements()) {
              Vehicle vehicle = (Vehicle) VehicleEnum.nextElement();

              fOut.writeInt(vehicle.fType);
              fOut.writeUTF(vehicle.fLicensePlate);
              fOut.writeFloat(vehicle.fSpeed);
              fOut.writeFloat(vehicle.fCurrentAcceleration);
              fOut.writeFloat(vehicle.fDistanceTraveled);
              fOut.writeInt(vehicle.fCurrentSegment);
              fOut.writeFloat(vehicle.fPositionOnCurrentSegment);
              fOut.writeInt(vehicle.fCurrentLane);
              fOut.writeBoolean(vehicle.fDesiredSpeedReached);
              fOut.writeFloat(vehicle.fDriver.fDesiredSpeed);
              fOut.writeFloat(vehicle.fDriver.fAggression);
              fOut.writeBoolean(vehicle.fDriver.fMiddleLaneDriver);
              fOut.writeBoolean(vehicle.fMandatoryLocationReached);
              fOut.writeBoolean(vehicle.fLeftBlinkerOn);
              fOut.writeBoolean(vehicle.fRightBlinkerOn);
            }
          }
        }

        // send the information the loopdetectors have gathered
        fOut.writeInt(fResources.fsimLoopDetectors.size());
        Enumeration LoopDetectorEnum = fResources.fsimLoopDetectors.elements();
        while (LoopDetectorEnum.hasMoreElements()) {
          LoopDetector loopDetector = (LoopDetector) LoopDetectorEnum.nextElement();

          fOut.writeUTF(loopDetector.fName);
          fOut.writeInt(loopDetector.fLocation.fID);
          fOut.writeInt(loopDetector.fTimeForMeasurement);
          fOut.writeInt(loopDetector.fVehiclesPassed);
          fOut.writeInt(loopDetector.fCarsPassed);
          fOut.writeInt(loopDetector.fTrucksPassed);
          fOut.writeFloat(loopDetector.fAverageSpeed);
          fOut.writeFloat(loopDetector.fAverageFlowRate);
          fOut.writeFloat(loopDetector.fAverageDensity);
        }

        // send the total number of vehicles in the network
        int TotalNrOfVehicles = 0;
        Enumeration LinkEnum = fResources.fNetwork.fLinks.elements();
        while (LinkEnum.hasMoreElements()) {
          Link link = (Link) LinkEnum.nextElement();

          TotalNrOfVehicles += link.fVehicleQueue.size();
        }
        fOut.writeInt(TotalNrOfVehicles);

        fOut.flush();
      }
      else if (Request == kDisconnect) {
        showMessage(fMessages.lookup("messageDisconnectReceived"));

        fOut.writeBoolean(true);
        fOut.flush();

        // get the date and time of disconnect
        Date CurrentDate = new Date();
        String Timestamp = DateFormat.getDateTimeInstance(DateFormat.SHORT,DateFormat.MEDIUM).format(CurrentDate);
        String ReportString = fMessages.lookup("messageLogDisconnect",
                                               Timestamp,
                                               String.valueOf(fIP),
                                               getProtocolAbbreviation());

        // log the disconnect
        fResources.fServerLog.addLine(ReportString);

        fIsOperational = false;
      }
      else {
        showMessage(fMessages.lookup("errorInvalidRequestReceived"));

        // indicate request-failure
        fOut.writeBoolean(false);
        fOut.flush();
      }
    }
    catch (IOException exc) {
      showMessage(fMessages.lookup("errorCommunicationFailure"));

      // get the date and time of disconnect
      Date CurrentDate = new Date();
      String Timestamp = DateFormat.getDateTimeInstance(DateFormat.SHORT,DateFormat.MEDIUM).format(CurrentDate);
      String ReportString = fMessages.lookup("messageLogDisconnect",
                                             Timestamp,
                                             String.valueOf(fIP),
                                             getProtocolAbbreviation());

      // log the disconnect
      fResources.fServerLog.addLine(ReportString);

      fIsOperational = false;
    }
  }

  public int getProtocol()
  {
    return kVisualizerProtocol;
  }

  public String getProtocolAbbreviation()
  {
    return fMessages.lookup("textVisualizerProtocolAbbreviation");;
  }

  public String translate(Messages messages)
  {
    return messages.lookup("textVisualizerProtocol");
  }
}
