// ----------------------------------
// Filename      : FileParser.java
// Created       : 19/12/1999
// Last modified : 10/03/2000
// Target        : Java VM
// Author        : Sven Maerivoet
//
// All rights reserved.
// Universitaire Instelling Antwerpen
// ----------------------------------

package mitrasim.util;

import java.io.*;
import mitrasim.exceptions.*;

/** @author Sven Maerivoet<BR>
    All rights reserved.<BR>
    Universitaire Instelling Antwerpen
*/
public class FileParser
{
  // internal datastructures
  private Messages         fMessages;
  private String           fFilename;
  private RandomAccessFile fFileToParse;
  public  int              fLineNr;
  public  String           fLineRead;

  public FileParser(Messages messages, String filename) throws GeneralError
  {
    fMessages = messages;
    fFilename = filename;
    try {
      // open the file
      fFileToParse = new RandomAccessFile(fFilename,"r");
      fLineNr = 0;
      fLineRead = "";
    }
    catch (IOException err) {
      // the specified file could not be opened
      throw (new GeneralError(fMessages.lookup("errorFileNotFound",filename),
                              GeneralError.kFileError));
    }
  }

  public void close()
  {
    try {
      fFileToParse.close();
    }
    catch (IOException err) {
    }
  }

  public String separatePrefix() throws GeneralError
  {
    if (fLineRead.length() == 0) {
      return "";
    }

    String Prefix = "";

    // construct the prefix (strings can be double-quoted to include whitespace)
    // by parsing until the first whitespace is encountered (or the end
    // of the string is reached)
    int Index = 0;
    boolean InString = false;
    if (fLineRead.length() > 0) {
      while ((Index < fLineRead.length()) &&
             (InString ||
              (!InString && (fLineRead.charAt(Index) != ' ') &&
                            (fLineRead.charAt(Index) != '\t')))) {
        if (fLineRead.charAt(Index) == '"') {
          InString = !InString;
        }
        else {
          if ((fLineRead.charAt(Index) == '\\') &&
              (Index == (fLineRead.length() - 1))) {
            // read a new line and stop parsing
            readNextLineFromFile();
            // construct a new prefix from the next line
            return separatePrefix();
          }
          else {
            Prefix = Prefix.concat(String.valueOf(fLineRead.charAt(Index)));
          }
        }
        ++Index;
      }
    }

    // remove the prefix from the string
    if (fLineRead.length() > 0) {
      fLineRead = fLineRead.substring(Index,fLineRead.length());
    }
    // remove leading and trailing whitespace
    fLineRead = fLineRead.trim();
    return Prefix;
  }

  public void readNextLineFromFile() throws GeneralError
  {
    try {
      boolean ContinueReading;

      do {
        // read a line from the file
        fLineRead = fFileToParse.readLine();
        if (fLineRead == null) {
          throw (new IOException());
        }
        ++fLineNr;

        // remove leading and trailing whitespace
        fLineRead = fLineRead.trim();

        // is it an empty line ?
        if (fLineRead.length() == 0) {
          // skip this line
          ContinueReading = true;
        }
        else {
          // is it a comment ?
          if (fLineRead.charAt(0) == '#') {
            // skip this line
            ContinueReading = true;
          }
          else {
            // find the optional comment
            int CommentPos = fLineRead.indexOf('#');
            if (CommentPos != -1) {
              // remove the optional comment
              fLineRead = fLineRead.substring(0,CommentPos);
              // remove leading and trailing whitespace
              fLineRead = fLineRead.trim();
              if (fLineRead.length() == 0) {
                // skip this line
                ContinueReading = true;
              }
              else  {
                // return this line
                ContinueReading = false;
              }
            }
            else  {
              // return this line
              ContinueReading = false;
            }
          }
        }
      }
      while (ContinueReading);
    }
    catch (IOException err) {
      // end-of-file reached
      throw (new GeneralError(fMessages.lookup("errorEOFTooSoonReached",String.valueOf(fLineNr)),
                              GeneralError.kParseError));
    }
  }

  public void readSectionHeader(String sectionName) throws GeneralError
  {
    readNextLineFromFile();

    // are we dealing with a section-identifier ?
    if (fLineRead.charAt(0) != '[') {
      throw (new GeneralError(fMessages.lookup("errorSectionExpected",sectionName,String.valueOf(fLineNr)),
                              GeneralError.kParseError));
    }

    // does the section-identifier ends with ']' ?
    if (fLineRead.charAt(fLineRead.length() - 1) != ']') {
      throw (new GeneralError(fMessages.lookup("errorCloseBracketExpected",String.valueOf(fLineNr)),
                              GeneralError.kParseError));
    }

    // extract the section-name
    if (fLineRead.length() > 2) {
      fLineRead = fLineRead.substring(1,fLineRead.length() - 1);
    }
    else {
      // no section-name was given
      fLineRead = "";
    }

    // is it the correct section-name ?
    if ((fLineRead.toUpperCase()).compareTo(sectionName.toUpperCase()) != 0) {
      throw (new GeneralError(fMessages.lookup("errorSectionExpected",sectionName,String.valueOf(fLineNr)),
                              GeneralError.kParseError));
    }

    readNextLineFromFile();

    // does the section-block begins with '{' ?
    if (fLineRead.charAt(0) != '{') {
      throw (new GeneralError(fMessages.lookup("errorOpenBraceExpected",String.valueOf(fLineNr)),
                              GeneralError.kParseError));
    }
  }

  public void checkForSectionClosure() throws GeneralError
  {
    // does the section-block ends with '}' ?
    if (fLineRead.charAt(0) != '}') {
      throw (new GeneralError(fMessages.lookup("errorCloseBraceExpected",String.valueOf(fLineNr)),
                              GeneralError.kParseError));
    }
  }
}
