// ----------------------------------
// Filename      : VisualizerGUI.java
// Created       : 22/04/2000
// Last modified : 25/06/2000
// Target        : Java VM
// Author        : Sven Maerivoet
//
// All rights reserved.
// Universitaire Instelling Antwerpen
// ----------------------------------

package mitrasim.visualizer;

import java.awt.*;
import java.awt.event.*;
import java.io.*;
import java.net.*;
import java.util.*;
import mitrasim.engine3d.*;
import mitrasim.exceptions.*;
import mitrasim.network.*;
import mitrasim.resources.*;
import mitrasim.server.*;
import mitrasim.simulator.*;
import mitrasim.util.*;

/** @author Sven Maerivoet<BR>
    All rights reserved.<BR>
    Universitaire Instelling Antwerpen
*/
public class VisualizerGUI extends Frame implements WindowListener, ActionListener, KeyListener, Runnable
{
  // internal datastructures
  private final boolean    kIgnoreDisconnectFailure     = true;
  private final boolean    kDontIgnoreDisconnectFailure = false;

  private Resources         fResources;
  private int               fWinXSize;
  private int               fWinYSize;
  private Messages          fMessages;
  private Network           fNetwork;
  private VisualizerCanvas  fVisualizerCanvas;
  private MessageCanvas     fMessageCanvas;
  private Button            fConnectButton;
  private Button            fDisconnectButton;
  private Button            fStartViewingButton;
  private Button            fStopViewingButton;
  private Button            fViewSpecificationButton;
  private Button            fSetFrameDelayButton;
  private MenuBar           fMenuBar;
  private String            fIPAddress;
  private int               fTCPPort;
  private boolean           fAboutBoxShown;
  private boolean           fConnected;
  private Socket            fSocket;
  private DataInputStream   fIn;
  private DataOutputStream  fOut;
  private boolean           fViewing;
  private Thread            fVisualizerThread;
  private Vector            fViewLinkList;
  private Vector            fViewEnvironmentList;
  private long              fTimeStamp;
  private boolean           fSimulatorRunning;
  private int               fMasterClock;
  private Renderer          fRenderer;
  private Hashtable         fLinkRenderingInformation;
  private int               fTotalNrOfVehicles;
  private boolean           fLogStatistics;
  private LoggingMenu       fLoggingMenu;
  private MenuItem          fStartLogging;
  private MenuItem          fStopLogging;
  private PrintWriter       fLogWriter;
  private final int         kWhitespace = 2;
  private boolean           fReverseKeyboardMode;
  private VLegend           fLegend;
  private ColorInterpolator fColorInterpolator;
  private CheckboxMenuItem  fMenuTrackingMode;

  public VisualizerGUI()
  {
    super();

    fResources = new Resources();
    fWinXSize = fResources.fvisXSize;
    fWinYSize = fResources.fvisYSize;
    fMessages = fResources.fMessages;
    fResources.fNetwork = new Network(fMessages);
    fNetwork = fResources.fNetwork;
    fIPAddress = "127.0.0.1";
    fTCPPort = 4242;
    fResources.fNetwork = new Network(fMessages);
    fViewLinkList = null;
    fViewEnvironmentList = null;
    fViewing = false;
    fTotalNrOfVehicles = 0;
    fLogStatistics = false;
    fReverseKeyboardMode = false;
    fResources.fvisShowTrafficFlow        = true;
    fResources.fvisAnimateVehicles        = true;
    fResources.fvisShowAverageSpeeds      = false;
    fResources.fvisShowDensities          = false;
    fResources.fvisShowVehicleInformation = false;
    fResources.fvisShowMoreVehicleDetail  = false;
    fColorInterpolator = new ColorInterpolator();

    // initialize the 3D-renderer
    fRenderer = new Renderer(-0.5f,-0.5f,0.5f,0.5f,0,0,0,0,null,0);

    // adjust the windowlayout
    fAboutBoxShown = false;
    setFont(new Font(fMessages.lookup("textFontName"),Font.PLAIN,14));
    setSize(fWinXSize,fWinYSize);
    setResizable(false);
    setLocation(0,0);
    setTitle(fMessages.lookup("textVisualizerWindowTitle"));
    setBackground(Color.black);
    setIconImage(Toolkit.getDefaultToolkit().getImage("M2000Icon.jpg"));

    setLayout(new BorderLayout());

    // create the panel containing the canvases
    Panel CanvasPanel = new Panel();
    CanvasPanel.setLayout(new BorderLayout());

    fVisualizerCanvas = new VisualizerCanvas(fWinXSize,fWinYSize - 120);
    CanvasPanel.add("North",fVisualizerCanvas);

    fMessageCanvas = new MessageCanvas(fWinXSize,40);
    CanvasPanel.add("South",fMessageCanvas);

    // create the panel containing the controls
    Panel ControlPanel = new Panel();
    ControlPanel.setLayout(new GridLayout(2,3));

    fConnectButton = new Button(fMessages.lookup("buttonConnect"));
    fConnectButton.addActionListener(this);
    fConnectButton.addKeyListener(this);
    ControlPanel.add(fConnectButton);

    fStartViewingButton = new Button(fMessages.lookup("buttonStartViewing"));
    fStartViewingButton.addActionListener(this);
    fStartViewingButton.addKeyListener(this);
    ControlPanel.add(fStartViewingButton);

    fViewSpecificationButton = new Button(fMessages.lookup("buttonViewSpecification"));
    fViewSpecificationButton.addActionListener(this);
    fViewSpecificationButton.addKeyListener(this);
    ControlPanel.add(fViewSpecificationButton);

    fDisconnectButton = new Button(fMessages.lookup("buttonDisconnect"));
    fDisconnectButton.addActionListener(this);
    fDisconnectButton.addKeyListener(this);
    ControlPanel.add(fDisconnectButton);

    fStopViewingButton = new Button(fMessages.lookup("buttonStopViewing"));
    fStopViewingButton.addActionListener(this);
    fStopViewingButton.addKeyListener(this);
    ControlPanel.add(fStopViewingButton);

    fSetFrameDelayButton = new Button(fMessages.lookup("buttonSetFrameDelay"));
    fSetFrameDelayButton.addActionListener(this);
    fSetFrameDelayButton.addKeyListener(this);
    ControlPanel.add(fSetFrameDelayButton);

    fDisconnectButton.setEnabled(false);
    fStartViewingButton.setEnabled(false);
    fStopViewingButton.setEnabled(false);
    fViewSpecificationButton.setEnabled(false);

    // add the panels
    add("North",CanvasPanel);
    add("South",ControlPanel);

    // create the menubar
    fMenuBar = new MenuBar();

    GeneralMenu generalMenu = new GeneralMenu(this,fMessages.lookup("menuGeneral"));
    fMenuBar.add(generalMenu);

    DisplayMenu displayMenu = new DisplayMenu(this,fMessages.lookup("menuDisplay"));
    fMenuBar.add(displayMenu);

    fLoggingMenu = new LoggingMenu(this,fMessages.lookup("menuLogging"));
    fMenuBar.add(fLoggingMenu);
    fLoggingMenu.setEnabled(false);

    setMenuBar(fMenuBar);

    // add the windowlistener
    addWindowListener(this);

    pack();
    show();

    if (fResources.fvisShowAboutBox) {
      // show the about-box
      AboutBox about = new AboutBox(this,"Visualizer");
    }
    fAboutBoxShown = true;

    fConnected = false;

    // create the visualizer's thread
    fVisualizerThread = new Thread(this);
    fVisualizerThread.start();

    // show the legend
    fLegend = new VLegend();
    fLegend.setVisible(true);
  }

  // methods from the WindowListener-interface
  public void windowOpened(WindowEvent e) { }
  public void windowClosing(WindowEvent e)
  {
    // disconnect from the server if necessary
    disconnect(kIgnoreDisconnectFailure);

    // quit the running program
    System.exit(0);
  }
  public void windowClosed(WindowEvent e) { }
  public void windowIconified(WindowEvent e) { }
  public void windowDeiconified(WindowEvent e) { }
  public void windowActivated(WindowEvent e) { }
  public void windowDeactivated(WindowEvent e) { }

  // method from the ActionListener-interface
  public void actionPerformed(ActionEvent e)
  {
    String Command = e.getActionCommand();

    if (Command.equals(fMessages.lookup("menuAbout"))) {
      AboutBox aboutBox = new AboutBox(this,"Visualizer");
    }

    if (Command.equals(fMessages.lookup("menuQuit"))) {
      windowClosing(null);
    }

    if (Command.equals(fMessages.lookup("menuVehicleToTrack"))) {

      VehicleTrackBox vehicleTrackBox = new VehicleTrackBox(this);

      if (!vehicleTrackBox.canceled()) {

        fResources.fvisTrackingMode = true;
        fMenuTrackingMode.setState(true);
      }
    }

    if (Command.equals(fMessages.lookup("menuStartLogging"))) {
      FileDialog fd = new FileDialog(this,
                                     fMessages.lookup("textSetLogFileDialogTitle"),
                                     FileDialog.SAVE);
      fd.show();
      String Dir = fd.getDirectory();
      String Filename = fd.getFile();

      if ((Dir != null) && (Filename != null)) {
        try {
          FileOutputStream LogFile = new FileOutputStream(Dir + Filename);
          fLogWriter = new PrintWriter(LogFile);

          // save a header
          fLogWriter.print(expand("Clocktick",8 + kWhitespace));
          fLogWriter.print(expand("hh",2 + kWhitespace));
          fLogWriter.print(expand("mm",2 + kWhitespace));
          fLogWriter.print(expand("ss",2 + kWhitespace));
          fLogWriter.print(expand("#veh.",7 + kWhitespace));

          for (int i = 0; i < fResources.fsimLoopDetectors.size(); ++i) {

            fLogWriter.print(expand("Loopdetector",40 + kWhitespace));
            fLogWriter.print(expand("#veh.",7 + kWhitespace));
            fLogWriter.print(expand("#cars",7 + kWhitespace));
            fLogWriter.print(expand("#trucks",7 + kWhitespace));
            fLogWriter.print(expand("<v>",3 + kWhitespace));
            fLogWriter.print(expand("<q>",7 + kWhitespace));
            fLogWriter.print(expand("<k>",7 + kWhitespace));
          }
          fLogWriter.println();
        }
        catch (IOException exc) {
          return;
        }

        fStartLogging.setEnabled(false);
        fStopLogging.setEnabled(true);

        fLogStatistics = true;
      }
    }

    if (Command.equals(fMessages.lookup("menuStopLogging"))) {

      fLogStatistics = false;

      fStartLogging.setEnabled(true);
      fStopLogging.setEnabled(false);

      fLogWriter.close();
    }

    if (Command.equals(fMessages.lookup("buttonConnect"))) {
      connect();
    }

    if (Command.equals(fMessages.lookup("buttonDisconnect"))) {
      disconnect(kDontIgnoreDisconnectFailure);
    }

    if (Command.equals(fMessages.lookup("buttonStartViewing"))) {
      fDisconnectButton.setEnabled(false);
      fStartViewingButton.setEnabled(false);
      fStopViewingButton.setEnabled(true);
      fViewSpecificationButton.setEnabled(false);
      fViewing = true;
    }

    if (Command.equals(fMessages.lookup("buttonStopViewing"))) {
      fDisconnectButton.setEnabled(true);
      fStartViewingButton.setEnabled(true);
      fStopViewingButton.setEnabled(false);
      fViewSpecificationButton.setEnabled(true);
      fViewing = false;
    }

    if (Command.equals(fMessages.lookup("buttonSetFrameDelay"))) {
      SetFrameDelayBox setFrameDelayBox = new SetFrameDelayBox(this);
    }

    if (Command.equals(fMessages.lookup("buttonViewSpecification"))) {
      setViewSpecification();
    }
  }

  // methods from the KeyListener-interface
  public void keyTyped(KeyEvent e) { }
  public void keyPressed(KeyEvent e)
  {
    int KeyCode = e.getKeyCode();

    if (KeyCode == KeyEvent.VK_LEFT) {
      if (!fReverseKeyboardMode) {
        fResources.fvisViewPointX -= fResources.kvisDistanceStepSize;
      }
      else {
        fResources.fvisViewPointX += fResources.kvisDistanceStepSize;
      }
      fVisualizerCanvas.repaint();
    }
    else if (KeyCode == KeyEvent.VK_RIGHT) {
      if (!fReverseKeyboardMode) {
        fResources.fvisViewPointX += fResources.kvisDistanceStepSize;
      }
      else {
        fResources.fvisViewPointX -= fResources.kvisDistanceStepSize;
      }
      fVisualizerCanvas.repaint();
    }
    else if (KeyCode == KeyEvent.VK_UP) {
      if (!fReverseKeyboardMode) {
        fResources.fvisViewPointY += fResources.kvisDistanceStepSize;
      }
      else {
        fResources.fvisViewPointY -= fResources.kvisDistanceStepSize;
      }
      fVisualizerCanvas.repaint();
    }
    else if (KeyCode == KeyEvent.VK_DOWN) {
      if (!fReverseKeyboardMode) {
        fResources.fvisViewPointY -= fResources.kvisDistanceStepSize;
      }
      else {
        fResources.fvisViewPointY += fResources.kvisDistanceStepSize;
      }
      fVisualizerCanvas.repaint();
    }
    else if (KeyCode == KeyEvent.VK_ADD) {
      decreaseDistance();
    }
    else if (KeyCode == KeyEvent.VK_SUBTRACT) {
      increaseDistance();
    }
    else if (KeyCode == KeyEvent.VK_NUMPAD4) {
      rotateZLeft();
    }
    else if (KeyCode == KeyEvent.VK_NUMPAD6) {
      rotateZRight();
    }
    else if (KeyCode == KeyEvent.VK_NUMPAD2) {
      rotateXYRight();
    }
    else if (KeyCode == KeyEvent.VK_NUMPAD8) {
      rotateXYLeft();
    }
    else if (KeyCode == KeyEvent.VK_R) {
      fReverseKeyboardMode = !fReverseKeyboardMode;
    }
  }
  public void keyReleased(KeyEvent e) { }

  private void rotateXYLeft()
  {
    fResources.fvisEyePointPhi -= fResources.kvisRotateStepSize;
    if (fResources.fvisEyePointPhi > 180) {
      fResources.fvisEyePointPhi = fResources.fvisEyePointPhi - 360;
    }
    if (fResources.fvisEyePointPhi < -180) {
      fResources.fvisEyePointPhi = fResources.fvisEyePointPhi + 360;
    }
    fVisualizerCanvas.repaint();
  }

  private void rotateXYRight()
  {
    fResources.fvisEyePointPhi += fResources.kvisRotateStepSize;
    if (fResources.fvisEyePointPhi > 180) {
      fResources.fvisEyePointPhi = fResources.fvisEyePointPhi - 360;
    }
    if (fResources.fvisEyePointPhi < -180) {
      fResources.fvisEyePointPhi = fResources.fvisEyePointPhi + 360;
    }
    fVisualizerCanvas.repaint();
  }

  private void rotateZLeft()
  {
    fResources.fvisEyePointTheta -= fResources.kvisRotateStepSize;
    if (fResources.fvisEyePointTheta > 360) {
      fResources.fvisEyePointTheta = fResources.fvisEyePointTheta - 360;
    }
    if (fResources.fvisEyePointTheta < 0) {
      fResources.fvisEyePointTheta = fResources.fvisEyePointTheta + 360;
    }
    fVisualizerCanvas.repaint();
  }

  private void rotateZRight()
  {
    fResources.fvisEyePointTheta += fResources.kvisRotateStepSize;
    if (fResources.fvisEyePointTheta > 360) {
      fResources.fvisEyePointTheta = fResources.fvisEyePointTheta - 360;
    }
    if (fResources.fvisEyePointTheta < 0) {
      fResources.fvisEyePointTheta = fResources.fvisEyePointTheta + 360;
    }
    fVisualizerCanvas.repaint();
  }

  private void increaseDistance()
  {
    fResources.fvisEyePointRho += fResources.kvisDistanceStepSize;
    fVisualizerCanvas.repaint();
  }

  private void decreaseDistance()
  {
    fResources.fvisEyePointRho -= fResources.kvisDistanceStepSize;
    if (fResources.fvisEyePointRho < 0) {
      fResources.fvisEyePointRho = 0;
    }
    fVisualizerCanvas.repaint();
  }

  private void connect()
  {
    try {
      ConnectBox connectBox = new ConnectBox(this,fIPAddress,fTCPPort);

      if (connectBox.canceled()) {
        return;
      }

      fMessageCanvas.showMessage(fMessages.lookup("messageConnectingToServer"));

      fIPAddress = connectBox.getIPAddress();
      fTCPPort = connectBox.getTCPPort();

      fSocket = new Socket(fIPAddress,fTCPPort);
      fIn = new DataInputStream(fSocket.getInputStream());
      fOut = new DataOutputStream(fSocket.getOutputStream());

      // acquire the server's identification-string
      String ServerIDString = fIn.readUTF();

      // request to use the remote control-protocol
      fOut.writeInt(Protocol.kVisualizerProtocol);
      fOut.flush();

      fMessageCanvas.clearMessage();

      // check if the server agrees
      int Indicator = fIn.readInt();

      if (Indicator != Protocol.kProtocolOk) {
        if (Indicator == Protocol.kInvalidProtocolRequested) {
          MessageBox ErrorBox = new MessageBox(this,
                                               fMessages.lookup("errorInvalidProtocolRequested"),
                                               MessageBox.kError);
        }
        disconnect(kDontIgnoreDisconnectFailure);
        return;
      }

      // download the network
      fOut.writeInt(VisualizerProtocol.kDownloadNetwork);
      fOut.flush();

      boolean Ok = fIn.readBoolean();
      if (!Ok) {
        MessageBox ErrorBox = new MessageBox(this,
                                             fMessages.lookup("errorNoNetworkAvailable"),
                                             MessageBox.kError);
        disconnect(kDontIgnoreDisconnectFailure);
        return;
      }

      fMessageCanvas.showMessage(fMessages.lookup("messageDownloadingNetwork"));

      fNetwork.readFromStream(fIn,Network.kDontShowTickMarks);
      fNetwork.readTimeStampFromStream(fIn);
      fTimeStamp = fNetwork.fTimeStamp;

      fMessageCanvas.showMessage(fMessages.lookup("messageCheckingNetwork"));

      // check the received network
      Report ResultReport = fResources.fNetwork.analyse();

      fMessageCanvas.clearMessage();

      if (!ResultReport.isEmpty()) {
        MessageBox ErrorBox = new MessageBox(this,
                                             fMessages.lookup("errorReceivedNetworkNotCompliant"),
                                             MessageBox.kError);
        disconnect(kDontIgnoreDisconnectFailure);
        return;
      }

      fMessageCanvas.showMessage(fMessages.lookup("messageCalculatingNetworkGeometry"));

      // calculate the network's geometry
      fNetwork.calcGeometry();

      fMessageCanvas.clearMessage();

      // download the view-specifications
      fMessageCanvas.showMessage(fMessages.lookup("messageDownloadingViewSpecifications"));

      // reserve room for 20 view-specifications and a load factor of 80%
      fResources.fvisViewSpecifications = new Hashtable(20,0.8f);

      int NrOfViewSpecifications = fIn.readInt();
      for (int i = 0; i < NrOfViewSpecifications; ++i) {
        ViewSpecification viewSpecification = new ViewSpecification();

        viewSpecification.fName = fIn.readUTF();
        viewSpecification.fPerspectiveMode = fIn.readBoolean();
        viewSpecification.fViewPointX = fIn.readInt();
        viewSpecification.fViewPointY = fIn.readInt();
        viewSpecification.fViewPointZ = fIn.readInt();
        viewSpecification.fEyePointRho = fIn.readInt();
        viewSpecification.fEyePointPhi = fIn.readInt();
        viewSpecification.fEyePointTheta = fIn.readInt();
        viewSpecification.fBoundingBoxMinX = fIn.readInt();
        viewSpecification.fBoundingBoxMinY = fIn.readInt();
        viewSpecification.fBoundingBoxMaxX = fIn.readInt();
        viewSpecification.fBoundingBoxMaxY = fIn.readInt();
        viewSpecification.fShowEnvironment = fIn.readBoolean();
        fResources.fvisViewSpecifications.put(new String(viewSpecification.fName),viewSpecification);
      }

      if (!fNetwork.fTitle.equalsIgnoreCase("")) {
        setTitle(fMessages.lookup("textVisualizerWindowTitle") +
                 " [" + fNetwork.fTitle + "]");
      }
      else {
        setTitle(fMessages.lookup("textVisualizerWindowTitle"));
      }

      fMessageCanvas.clearMessage();

      // reserve room for 20 loopdetectors and a load factor of 80%
      fResources.fsimLoopDetectors = new Hashtable(20,0.8f);

      fConnected = true;

      fConnectButton.setEnabled(false);
      fDisconnectButton.setEnabled(true);
      fViewSpecificationButton.setEnabled(true);
      fLoggingMenu.setEnabled(true);
      fStartLogging.setEnabled(true);
      fStopLogging.setEnabled(false);
    }
    catch (IOException exc) {
      fMessageCanvas.clearMessage();
      MessageBox ErrorBox = new MessageBox(this,
                                           fMessages.lookup("errorFailedToConnectWithServer"),
                                           MessageBox.kError);
    }
  }

  private void disconnect(boolean IgnoreFailure)
  {
    fConnected = false;
    fViewing = false;

    if (fOut == null) {
      return;
    }

    try {
      fMessageCanvas.showMessage(fMessages.lookup("messageDisconnecting"));

      fOut.writeInt(VisualizerProtocol.kDisconnect);
      fOut.flush();
      boolean Ok = fIn.readBoolean();

      fMessageCanvas.clearMessage();

      if (!Ok) {
        throw (new IOException());
      }

      fSocket.close();
    }
    catch (IOException exc) {
      fMessageCanvas.clearMessage();
      if (!IgnoreFailure) {
        MessageBox ErrorBox = new MessageBox(this,
                                             fMessages.lookup("errorFailedToDisconnectFromServer"),
                                             MessageBox.kError);
      }
    }

    fMessageCanvas.clearMessage();
    fVisualizerCanvas.repaint();

    fConnectButton.setEnabled(true);
    fDisconnectButton.setEnabled(false);
    fStartViewingButton.setEnabled(false);
    fStopViewingButton.setEnabled(false);
    fViewSpecificationButton.setEnabled(false);
    fStartLogging.setEnabled(false);
    fStopLogging.setEnabled(false);
    fLoggingMenu.setEnabled(false);
  }

  private void setViewSpecification()
  {
    ViewSpecificationBox viewSpecificationBox = new ViewSpecificationBox(this);

    if (!viewSpecificationBox.canceled()) {
      BoundingBox SpecifiedBoundingBox = new BoundingBox();
      SpecifiedBoundingBox.fMinX = fResources.fvisBoundingBoxMinX;
      SpecifiedBoundingBox.fMinY = fResources.fvisBoundingBoxMinY;
      SpecifiedBoundingBox.fMaxX = fResources.fvisBoundingBoxMaxX;
      SpecifiedBoundingBox.fMaxY = fResources.fvisBoundingBoxMaxY;

      // build viewlists by including all links and environments that
      // are completely contained in the specified bounding box

      fViewLinkList = new Vector();
      Enumeration LinkEnum = fNetwork.fLinks.elements();
      while (LinkEnum.hasMoreElements()) {
        Link link = (Link) LinkEnum.nextElement();

        // is the link completely contained in the specified bounding box ?
        // check if all the nodes (in the link) lie in the specified bounding box
        boolean FullyContained = true;
        for (int i = 0; i < link.fNrOfNodes; ++i) {
          if (!SpecifiedBoundingBox.contains2D(link.fNodes[i].fXPos,
                                               link.fNodes[i].fYPos)) {
            FullyContained = false;
          }
        }

        if (FullyContained) {
          // add the link to the viewlist
          fViewLinkList.addElement(link);
        }
      }

      fViewEnvironmentList = new Vector();
      Enumeration EnvironmentEnum = fNetwork.fEnvironment.elements();
      while (EnvironmentEnum.hasMoreElements()) {
        Environment environment = (Environment) EnvironmentEnum.nextElement();

        // is the environment completely contained in the specified bounding box ?
        // check if all the nodes (in the environment) lie in the specified bounding box
        boolean FullyContained = true;
        for (int i = 0; i < environment.fNrOfNodes; ++i) {
          if (!SpecifiedBoundingBox.contains2D(environment.fNodes[i].fXPos,
                                               environment.fNodes[i].fYPos)) {
            FullyContained = false;
          }
        }

        if (FullyContained) {
          // add the environment to the viewlist
          fViewEnvironmentList.addElement(environment);
        }
      }

      if (fViewLinkList.size() == 0) {
        fStartViewingButton.setEnabled(false);
        fStopViewingButton.setEnabled(false);
        MessageBox ErrorBox = new MessageBox(this,
                                             fMessages.lookup("errorNoLinksInBoundingBox"),
                                             MessageBox.kError);
      }
      else {

        try {
          fMessageCanvas.showMessage(fMessages.lookup("messageSendingActiveLinks"));

          // transfer the requested links to the server
          fOut.writeInt(VisualizerProtocol.kSetActiveLinks);

          fOut.writeInt(fViewLinkList.size());

          Enumeration ViewLinkEnum = fViewLinkList.elements();
          while (ViewLinkEnum.hasMoreElements()) {
            Link link = (Link) ViewLinkEnum.nextElement();
            fOut.writeInt(link.fID);
          }

          fOut.flush();

          fMessageCanvas.clearMessage();

          fLinkRenderingInformation = new Hashtable(200,0.8f);

          ViewLinkEnum = fViewLinkList.elements();
          while (ViewLinkEnum.hasMoreElements()) {
            Link link = (Link) ViewLinkEnum.nextElement();

            LinkRenderingInformation lri = new LinkRenderingInformation();
            lri.fLeft = new Point3D[link.fNrOfNodes];
            lri.fRight = new Point3D[link.fNrOfNodes];
            lri.fLanePositions = new Point3D[link.fNrOfNodes][];

            for (int NodeNr = 0; NodeNr < link.fNrOfNodes; ++NodeNr) {

              lri.fLeft[NodeNr] = new Point3D();
              lri.fRight[NodeNr] = new Point3D();

              // retrieve the number of lanes the link has at the specified node
              int NrOfLanesAtNode = link.fNrOfLanesAtNode[NodeNr];

              // calculate the width of the road at the current node
              float HalfRoadWidth = (NrOfLanesAtNode * fNetwork.fLaneWidth) / 2.0f;

              Point3D p = new Point3D();
              Point3D q = new Point3D();

              if (NodeNr < (link.fNrOfNodes - 1)) {
                p.fX = link.fNodes[NodeNr].fXPos;
                p.fY = link.fNodes[NodeNr].fYPos;
                p.fZ = link.fNodes[NodeNr].fZPos;
                q.fX = link.fNodes[NodeNr + 1].fXPos;
                q.fY = link.fNodes[NodeNr + 1].fYPos;
                q.fZ = link.fNodes[NodeNr + 1].fZPos;
                lri.fLeft[NodeNr] = fRenderer.addPointByLine(p,q,HalfRoadWidth,(float) (Math.PI / 2.0));
                lri.fRight[NodeNr] = fRenderer.addPointByLine(p,q,HalfRoadWidth,(float) -(Math.PI / 2.0));
              }
              else {
                // handle the last node
                p.fX = link.fNodes[NodeNr].fXPos;
                p.fY = link.fNodes[NodeNr].fYPos;
                p.fZ = link.fNodes[NodeNr].fZPos;
                q.fX = link.fNodes[NodeNr - 1].fXPos;
                q.fY = link.fNodes[NodeNr - 1].fYPos;
                q.fZ = link.fNodes[NodeNr - 1].fZPos;
                lri.fLeft[NodeNr] = fRenderer.addPointByLine(p,q,HalfRoadWidth,(float) -(Math.PI / 2.0));
                lri.fRight[NodeNr] = fRenderer.addPointByLine(p,q,HalfRoadWidth,(float) (Math.PI / 2.0));
              }

              lri.fLanePositions[NodeNr] = new Point3D[NrOfLanesAtNode - 1];
              for (int LaneNr = 1; LaneNr <= (NrOfLanesAtNode - 1); ++LaneNr) {
                lri.fLanePositions[NodeNr][LaneNr - 1] = new Point3D();
                lri.fLanePositions[NodeNr][LaneNr - 1] = fRenderer.addPointByLine(lri.fRight[NodeNr],
                                                                                  lri.fLeft[NodeNr],
                                                                                  LaneNr * fNetwork.fLaneWidth,
                                                                                  0.0f);
              }
            }

            fLinkRenderingInformation.put(new Integer(link.fID),lri);
          }

          // automatically start viewing
          fDisconnectButton.setEnabled(false);
          fStartViewingButton.setEnabled(false);
          fStopViewingButton.setEnabled(true);
          fViewSpecificationButton.setEnabled(false);
          fViewing = true;
        }
        catch (IOException exc) {
          fMessageCanvas.clearMessage();
          MessageBox ErrorBox = new MessageBox(this,
                                               fMessages.lookup("errorCommunicationFailure"),
                                               MessageBox.kError);
          disconnect(kIgnoreDisconnectFailure);
        }
      }
    }
  }

  public void run()
  {
    while (true) {
      if (fViewing && fConnected) {

        try {
          fOut.writeInt(VisualizerProtocol.kRequestData);
          fOut.flush();

          fResources.fNetwork.readTimeStampFromStream(fIn);
          // check if server and visualizer are still talking about the same network
          if ((fTimeStamp != fResources.fNetwork.fTimeStamp) && fConnected) {
            fViewing = false;
            MessageBox ErrorBox = new MessageBox(this,
                                                 fMessages.lookup("errorNetworkHasChanged"),
                                                 MessageBox.kError);
            disconnect(kIgnoreDisconnectFailure);
          }
          else {

            // request the state of the simulator
            fSimulatorRunning = fIn.readBoolean();

            // retrieve the current value of the simulator's masterclock
            fMasterClock = fIn.readInt();

            // request the data on the vehicles of all the links
            Enumeration ViewLinkEnum = fViewLinkList.elements();
            while (ViewLinkEnum.hasMoreElements()) {
              Link link = (Link) ViewLinkEnum.nextElement();

              link.fVehicleQueue = new Vector();

              link.fSegmentAverageSpeed = new float[link.fNrOfNodes - 1];
              link.fSegmentDensity = new float[link.fNrOfNodes - 1];
              int[] SegmentNrOfVehicles = new int[link.fNrOfNodes - 1];

              for (int i = 0; i < (link.fNrOfNodes - 1); ++i) {
                link.fSegmentAverageSpeed[i] = 0.0f;
                link.fSegmentDensity[i] = 0.0f;
                SegmentNrOfVehicles[i] = 0;
              }

              int NrOfVehicles = fIn.readInt();
              for (int i = 0; i < NrOfVehicles; ++i) {
                Vehicle vehicle = new Vehicle();

                vehicle.fType = fIn.readInt();
                vehicle.fLicensePlate = fIn.readUTF();
                vehicle.fSpeed = fIn.readFloat();
                vehicle.fCurrentAcceleration = fIn.readFloat();
                vehicle.fDistanceTraveled = fIn.readFloat();
                vehicle.fCurrentSegment = fIn.readInt();
                vehicle.fPositionOnCurrentSegment = fIn.readFloat();
                vehicle.fCurrentLane = fIn.readInt();
                vehicle.fDesiredSpeedReached = fIn.readBoolean();
                vehicle.fDriver = new Driver();
                vehicle.fDriver.fDesiredSpeed = fIn.readFloat();
                vehicle.fDriver.fAggression = fIn.readFloat();
                vehicle.fDriver.fMiddleLaneDriver = fIn.readBoolean();
                vehicle.fMandatoryLocationReached = fIn.readBoolean();
                vehicle.fLeftBlinkerOn = fIn.readBoolean();
                vehicle.fRightBlinkerOn = fIn.readBoolean();

                if (vehicle.fType == Vehicle.kCar) {
                  vehicle.fLength = fResources.ksimCarLength;
                  vehicle.fWidth = fResources.ksimCarWidth;
                  vehicle.fHeight = fResources.ksimCarHeight;
                }
                else {
                  vehicle.fLength = fResources.ksimTruckLength;
                  vehicle.fWidth = fResources.ksimTruckWidth;
                  vehicle.fHeight = fResources.ksimTruckHeight;
                }

                // add vehicle to link's vehiclequeue
                link.fVehicleQueue.addElement(vehicle);

                link.fSegmentAverageSpeed[vehicle.fCurrentSegment] += vehicle.fSpeed;
                ++link.fSegmentDensity[vehicle.fCurrentSegment];
                ++SegmentNrOfVehicles[vehicle.fCurrentSegment];
              }

              // calculate the average speed in each segment
              for (int i = 0; i < (link.fNrOfNodes - 1); ++i) {

                if (SegmentNrOfVehicles[i] != 0) {
                  link.fSegmentAverageSpeed[i] = (link.fSegmentAverageSpeed[i] /
                                                  SegmentNrOfVehicles[i]);
                }
                else {
                  link.fSegmentAverageSpeed[i] = link.fSpeedLimit;
                }
              }

              // calculate the density in each segment
              for (int i = 0; i < (link.fNrOfNodes - 1); ++i) {
                // the quick way
                link.fSegmentDensity[i] = ((link.fSegmentDensity[i] /
                                            link.fLengths[i]) * 1000.0f) /
                                          link.fNrOfLanesAtNode[i];
              }

            }

            // request the data gathered by the network's loopdetectors
            fResources.fsimLoopDetectors.clear();
            int NrOfLoopDetectors = fIn.readInt();

            for (int i = 0; i < NrOfLoopDetectors; ++i) {

              String Name = fIn.readUTF();
              int NodeID = fIn.readInt();
              int TimeForMeasurement = fIn.readInt();
              int VehiclesPassed = fIn.readInt();
              int CarsPassed = fIn.readInt();
              int TrucksPassed = fIn.readInt();
              float AverageSpeed = fIn.readFloat();
              float AverageFlowRate = fIn.readFloat();
              float AverageDensity = fIn.readFloat();

              LoopDetector loopDetector = new LoopDetector(Name,null,TimeForMeasurement);
              loopDetector.fVehiclesPassed = VehiclesPassed;
              loopDetector.fCarsPassed = CarsPassed;
              loopDetector.fTrucksPassed = TrucksPassed;
              loopDetector.fAverageSpeed = AverageSpeed;
              loopDetector.fAverageFlowRate = AverageFlowRate;
              loopDetector.fAverageDensity = AverageDensity;

              fResources.fsimLoopDetectors.put(new Integer(NodeID),loopDetector);
            }

            // receive the total number of vehicles in the simulator
            fTotalNrOfVehicles = fIn.readInt();

            // redraw the screen
            fVisualizerCanvas.repaint();

            // wait for the framedelay
            try {
              Thread.sleep(fResources.fvisFrameDelay);
            }
            catch (InterruptedException exc) {
            }

            // if necessary, dump data to logfile
            if (fLogStatistics) {

              fLogWriter.print(expand(String.valueOf(fMasterClock),8 + kWhitespace));

              int Clockticks = fMasterClock;
              // filter out the days when a 24-hour cycle has completed
              Clockticks = Clockticks % (10 * 60 * 60 * 24);

              // calculate the hours, minutes and seconds
              int Hours = Clockticks / (10 * 60 * 60);
              Clockticks = Clockticks % (10 * 60 * 60);
              int Minutes = Clockticks / (10 * 60);
              Clockticks = Clockticks % (10 * 60);
              int Seconds = Clockticks / 10;
              Clockticks = Clockticks % 10;

              fLogWriter.print(expand(String.valueOf(Hours),2 + kWhitespace));
              fLogWriter.print(expand(String.valueOf(Minutes),2 + kWhitespace));
              fLogWriter.print(expand(String.valueOf(Seconds),2 + kWhitespace));

              fLogWriter.print(expand(String.valueOf(fTotalNrOfVehicles),7 + kWhitespace));

              Enumeration LoopDetectorEnum = fResources.fsimLoopDetectors.elements();
              while (LoopDetectorEnum.hasMoreElements()) {
                 LoopDetector loopDetector = (LoopDetector) LoopDetectorEnum.nextElement();

                 fLogWriter.print(expand(loopDetector.fName,40 + kWhitespace));
                 fLogWriter.print(expand(String.valueOf(loopDetector.fVehiclesPassed),7 + kWhitespace));
                 fLogWriter.print(expand(String.valueOf(loopDetector.fCarsPassed),7 + kWhitespace));
                 fLogWriter.print(expand(String.valueOf(loopDetector.fTrucksPassed),7 + kWhitespace));
                 fLogWriter.print(expand(String.valueOf(Math.round(loopDetector.fAverageSpeed)),3 + kWhitespace));
                 fLogWriter.print(expand(String.valueOf(Math.round(loopDetector.fAverageFlowRate)),7 + kWhitespace));
                 fLogWriter.print(expand(String.valueOf(Math.round(loopDetector.fAverageDensity)),7 + kWhitespace));
              }

              fLogWriter.println();
            }
          }
        }
        catch (IOException exc) {
          fViewing = false;
          MessageBox ErrorBox = new MessageBox(this,
                                               fMessages.lookup("errorCommunicationFailure"),
                                               MessageBox.kError);
          disconnect(kIgnoreDisconnectFailure);
        }
      }

      Thread.yield();
    }
  }

  private String expand(String string, int width)
  {
    String Buffer = string;
    for (int i = 0; i < (width - string.length()); ++i) {
      Buffer = Buffer.concat(" ");
    }
    return Buffer;
  }

  /** @author Sven Maerivoet<BR>
      All rights reserved.<BR>
      Universitaire Instelling Antwerpen
  */  
  // the visualizer's canvas
  private class VisualizerCanvas extends Canvas
  {
    // internal datastructures
    private final int k2DPixelBorder = 25;

    private int   cx1;
    private int   cy1;
    private int   cx2;
    private int   cy2;
    private int   cwidth;
    private int   cheight;
    private Image fCanvasImage;

    public VisualizerCanvas(int width, int height)
    {
      super();

      setSize(width,height);

      repaint();
    }

    public void update(Graphics g)
    {
      paint(g);
    }

    public boolean imageUpdate(Image img, int flags, int x, int y, int w, int h)
    {
      // only allow repainting of this canvas when the image is completely loaded
      // this prevents flickering
      if ((flags & ALLBITS) != 0) {
        repaint();
      }
      return ((flags & (ALLBITS|ERROR)) == 0);
    }

    private void convertNode(Node node, Point3D point)
    {
      point.set(node.fXPos,node.fYPos,node.fZPos);
    }

    private double sqr(double x)
    {
      return (x * x);
    }

    public void paint(Graphics cg)
    {
      // define the boundaries for drawing on the canvas
      cx1 = 10;
      cy1 = 10;
      cx2 = (cx1 + getSize().width) - (getInsets().right + 10 + 10);
      cy2 = (cy1 + getSize().height) - (getInsets().bottom + 10 + 10);
      cwidth = cx2 - cx1;
      cheight = cy2 - cy1;

      // define the boundaries for drawing in the image
      int ix1 = 0;
      int iy1 = 0;
      int ix2 = cwidth - 1;
      int iy2 = cheight - 1;
      int iwidth = ix2 - ix1;
      int iheight = iy2 - iy1;

      fCanvasImage = createImage(iwidth,iheight);
      Graphics g = fCanvasImage.getGraphics();

      // draw a border on the canvas
      cg.setColor(fResources.kvisBorderColor);
      cg.drawRoundRect(cx1 - 1,cy1 - 1,cwidth + 2,cheight + 2,10,10);

      // clear the draw-area
      g.clearRect(ix1,iy1,iwidth + 1,iheight + 1);
      g.clipRect(ix1,iy1,iwidth + 1,iheight + 1);

      if (!fViewing) {
        // draw the image on the canvas
        cg.drawImage(fCanvasImage,cx1 + 1,cy1 + 1,this);

        return;
      }

      // configure the 3D-renderer
      fRenderer.setWVMapping(-0.5f,-0.5f,0.5f,0.5f,ix1,iy1,ix2,iy2);
      fRenderer.setYOffset(fWinYSize);
      fRenderer.setViewPoint(fResources.fvisViewPointX,
                             fResources.fvisViewPointY,
                             fResources.fvisViewPointZ);
      fRenderer.setEyePoint(fResources.fvisEyePointRho,
                            fRenderer.convertDegToRad(fResources.fvisEyePointPhi),
                            fRenderer.convertDegToRad(fResources.fvisEyePointTheta));
      fRenderer.setGraphics(g);
      if (fResources.fvisPerspectiveMode) {
        fRenderer.setPerspectiveMode();
      }
      else {
        fRenderer.setParallelMode();
      }

      if (fResources.fvisShowTrafficFlow) {

        if (fResources.fvisShowEnvironment) {
  
          // show the specified environment
          Enumeration ViewEnvironmentEnum = fViewEnvironmentList.elements();
          while (ViewEnvironmentEnum.hasMoreElements()) {
            Environment environment = (Environment) ViewEnvironmentEnum.nextElement();

            // show environment in 3D
            g.setColor(environment.fColor);

            if (environment.fType == Environment.kRectangle) {

              Point3D Point[] = new Point3D[4];

              for (int i = 0; i < 4; ++i) {
                Point[i] = new Point3D();
              }

              Point[0].fX = environment.fNodes[0].fXPos;
              Point[0].fY = environment.fNodes[0].fYPos;
              Point[0].fZ = environment.fNodes[0].fZPos;

              Point[1].fX = environment.fNodes[0].fXPos;
              Point[1].fY = environment.fNodes[1].fYPos;
              Point[1].fZ = environment.fNodes[0].fZPos;

              Point[2].fX = environment.fNodes[1].fXPos;
              Point[2].fY = environment.fNodes[1].fYPos;
              Point[2].fZ = environment.fNodes[0].fZPos;

              Point[3].fX = environment.fNodes[1].fXPos;
              Point[3].fY = environment.fNodes[0].fYPos;
              Point[3].fZ = environment.fNodes[0].fZPos;

              if (environment.fFilled) {
                Polygon Rectangle = new Polygon();
                for (int i = 0; i < 4; ++i) {
                  ViewportPoint Pos = fRenderer.renderPoint(Point[i]);
                  Rectangle.addPoint(Pos.fX,Pos.fY);
                }
                g.fillPolygon(Rectangle);
              }
              else {
                // draw the outline
                fRenderer.plotLine(Point[0],Point[1]);
                fRenderer.plotLine(Point[1],Point[2]);
                fRenderer.plotLine(Point[2],Point[3]);
                fRenderer.plotLine(Point[3],Point[0]);
              }
            }
            else if (environment.fType == Environment.kPolygon) {

              Point3D Point = new Point3D();
              Polygon polygon = new Polygon();
              ViewportPoint Pos = new ViewportPoint();

              for (int i = 0; i < environment.fNrOfNodes; ++i) {
                convertNode(environment.fNodes[i],Point);
                Pos = fRenderer.renderPoint(Point);
                polygon.addPoint(Pos.fX,Pos.fY);
              }

              if (environment.fFilled) {
                g.fillPolygon(polygon);
              }
              else {
                // draw the outline
                g.drawPolygon(polygon);
              }
            }
            else if (environment.fType == Environment.kOval) {

              Point3D Point = new Point3D();
              Polygon polygon = new Polygon();
              ViewportPoint Pos = new ViewportPoint();

              for (int i = 0; i < fResources.fnvOvalInterpolation; ++i) {
                float Angle = (float) (((2.0 * Math.PI) / fResources.fnvOvalInterpolation) * i);
                Point.fX = environment.fNodes[0].fXPos + Math.round((float) environment.fXRadius * Math.cos(Angle));
                Point.fY = environment.fNodes[0].fYPos + Math.round((float) environment.fYRadius * Math.sin(Angle));
                Point.fZ = environment.fNodes[0].fZPos;
                Pos = fRenderer.renderPoint(Point);
                polygon.addPoint(Pos.fX,Pos.fY);
              }

              if (environment.fFilled) {
                g.fillPolygon(polygon);
              }
              else {
                // draw the outline
                g.drawPolygon(polygon);
              }
            }
            else if (environment.fType == Environment.kRailroad) {

              Point3D p1 = new Point3D();
              Point3D p2 = new Point3D();

              // draw straight-lines between the linksegments
              for (int i = 0; i < (environment.fNrOfNodes - 1); ++i) {

                convertNode(environment.fNodes[i],p1);
                convertNode(environment.fNodes[i + 1],p2);

                // calculate the extent of the railroadtrack
                // (the points lie perpendicular to the direction of the track)
                float Extent = (fNetwork.fRailroadWidth / 2.0f);

                Point3D p1l = fRenderer.addPointByLine(p1,p2,Extent,(float) (Math.PI / 2.0));
                Point3D p1r = fRenderer.addPointByLine(p1,p2,Extent,(float) -(Math.PI / 2.0));
                Point3D p2l = fRenderer.addPointByLine(p2,p1,Extent,(float) -(Math.PI / 2.0));
                Point3D p2r = fRenderer.addPointByLine(p2,p1,Extent,(float) (Math.PI / 2.0));

                ViewportPoint vp1l = fRenderer.renderPoint(p1l);
                ViewportPoint vp1r = fRenderer.renderPoint(p1r);
                ViewportPoint vp2l = fRenderer.renderPoint(p2l);
                ViewportPoint vp2r = fRenderer.renderPoint(p2r);

                // draw the supporting sleepers
                float RailroadLength = (float) Math.sqrt(sqr(p2.fX - p1.fX) +
                                                         sqr(p2.fY - p1.fY) +
                                                         sqr(p2.fZ - p1.fZ));

                int NrOfSleepers = (int) Math.round(RailroadLength /
                                                    (fResources.kvisSleeperWidth +
                                                     fResources.kvisSleeperDistance ));

                g.setColor(fResources.kvisSleeperColor);

                for (int Sleeper = 0; Sleeper < NrOfSleepers; ++Sleeper) {
                  // determine the position of the sleeper using a parametric equation
                  float t = Sleeper * (1.0f / NrOfSleepers);

                  Point3D p = new Point3D();
                  p.fX = p1.fX + ((p2.fX - p1.fX) * t);
                  p.fY = p1.fY + ((p2.fY - p1.fY) * t);
                  p.fZ = p1.fZ + ((p2.fZ - p1.fZ) * t);

                  Point3D q = fRenderer.addPointByLine(p,p2,fResources.kvisSleeperWidth,0.0f);

                  float HalfWidth = fResources.kvisSleeperLength / 2.0f;

                  Point3D upperleft = fRenderer.addPointByLine(p,p2,HalfWidth,(float) (Math.PI / 2.0));
                  Point3D upperright = fRenderer.addPointByLine(q,p2,HalfWidth,(float) (Math.PI / 2.0));
                  Point3D lowerright = fRenderer.addPointByLine(q,p2,HalfWidth,(float) -(Math.PI / 2.0));
                  Point3D lowerleft = fRenderer.addPointByLine(p,p2,HalfWidth,(float) -(Math.PI / 2.0));

                  ViewportPoint vpul = fRenderer.renderPoint(upperleft);
                  ViewportPoint vpur = fRenderer.renderPoint(upperright);
                  ViewportPoint vplr = fRenderer.renderPoint(lowerright);
                  ViewportPoint vpll = fRenderer.renderPoint(lowerleft);

                  Polygon SleeperBar = new Polygon();
                  SleeperBar.addPoint(vpul.fX,vpul.fY);
                  SleeperBar.addPoint(vpur.fX,vpur.fY);
                  SleeperBar.addPoint(vplr.fX,vplr.fY);
                  SleeperBar.addPoint(vpll.fX,vpll.fY);
                  g.fillPolygon(SleeperBar);
                }

                // draw the rails
                g.setColor(fResources.kvisRailroadColor);
                g.drawLine(vp1l.fX,vp1l.fY,vp2l.fX,vp2l.fY);
                g.drawLine(vp1r.fX,vp1r.fY,vp2r.fX,vp2r.fY);
              }
            }
          }
        }

        // show the specified links
        Enumeration ViewLinkEnum = fViewLinkList.elements();
        while (ViewLinkEnum.hasMoreElements()) {
          Link link = (Link) ViewLinkEnum.nextElement();

          // show link in 3D
          LinkRenderingInformation lri = (LinkRenderingInformation)
                                         fLinkRenderingInformation.get(new Integer(link.fID));

          // handle each linksegment separately
          for (int NodeNr = 0; NodeNr < (link.fNrOfNodes - 1); ++NodeNr) {

            ViewportPoint vp1l = fRenderer.renderPoint(lri.fLeft[NodeNr]);
            ViewportPoint vp1r = fRenderer.renderPoint(lri.fRight[NodeNr]);
            ViewportPoint vp2l = fRenderer.renderPoint(lri.fLeft[NodeNr + 1]);
            ViewportPoint vp2r = fRenderer.renderPoint(lri.fRight[NodeNr + 1]);

            // draw the roadsurface
            if (fResources.fvisAnimateVehicles) {
              g.setColor(fResources.kvisRoadSurfaceColor);
            }
            else {

              if (fResources.fvisShowAverageSpeeds) {

                g.setColor(fColorInterpolator.interpolate((int) Math.round(link.fSegmentAverageSpeed[NodeNr])));
              }
              else if (fResources.fvisShowDensities) {

                g.setColor(fColorInterpolator.interpolate((int) Math.round(100.0f - link.fSegmentDensity[NodeNr])));
              }
            }

            Polygon RoadSurface = new Polygon();
            RoadSurface.addPoint(vp1l.fX,vp1l.fY);
            RoadSurface.addPoint(vp2l.fX,vp2l.fY);
            RoadSurface.addPoint(vp2r.fX,vp2r.fY);
            RoadSurface.addPoint(vp1r.fX,vp1r.fY);
            g.fillPolygon(RoadSurface);

            if (fResources.fvisAnimateVehicles) {

              // draw the road-extent
              g.setColor(fResources.kvisRoadExtentColor);
              g.drawLine(vp1l.fX,vp1l.fY,vp2l.fX,vp2l.fY);
              g.drawLine(vp1r.fX,vp1r.fY,vp2r.fX,vp2r.fY);

              // mark the bends and junctions
              g.setColor(fResources.kvisMarkBendJunctionColor);
              g.drawLine(vp1l.fX,vp1l.fY,vp1r.fX,vp1r.fY);
              g.drawLine(vp2l.fX,vp2l.fY,vp2r.fX,vp2r.fY);

              // mark the lanes
              g.setColor(fResources.kvisLaneSeparationColor);

              if (lri.fLanePositions[NodeNr].length == lri.fLanePositions[NodeNr + 1].length) {
                // draw marks between corresponding lane-positions
                for (int LaneNr = 1; LaneNr <= lri.fLanePositions[NodeNr].length; ++LaneNr) {
                  fRenderer.plotLine(lri.fLanePositions[NodeNr][LaneNr - 1],
                                     lri.fLanePositions[NodeNr + 1][LaneNr - 1]);
                }
              }
              else {
                // is the section of the road we are dealing with narrowed or widened ?

                if (lri.fLanePositions[NodeNr].length > lri.fLanePositions[NodeNr + 1].length) {
                  // this section of the road is narrowed

                  if (link.fLocation == Link.kLeft) {
                    // draw marks between corresponding lane-positions
                    for (int LaneNr = 1; LaneNr <= lri.fLanePositions[NodeNr + 1].length; ++LaneNr) {
                      fRenderer.plotLine(lri.fLanePositions[NodeNr][LaneNr - 1],
                                         lri.fLanePositions[NodeNr + 1][LaneNr - 1]);
                    }
                    // connect the last lane
                    g.setColor(fResources.kvisRampSeparationColor);
                    int LastLane = lri.fLanePositions[NodeNr + 1].length + 1;
                    fRenderer.plotLine(lri.fLanePositions[NodeNr][LastLane - 1],
                                       lri.fLeft[NodeNr + 1]);
                    // draw the marks of the merging lanes
                    g.setColor(fResources.kvisLaneSeparationColor);
                    for (int LaneNr = (LastLane + 1); LaneNr <= lri.fLanePositions[NodeNr].length; ++LaneNr) {
                      fRenderer.plotLine(lri.fLanePositions[NodeNr][LaneNr - 1],
                                         lri.fLeft[NodeNr + 1]);
                    }
                  }
                  else { // location is right
                    // draw the marks of the merging lanes
                    int NrOfMergingLanes = lri.fLanePositions[NodeNr].length -
                                           lri.fLanePositions[NodeNr + 1].length;
                    for (int LaneNr = 1; LaneNr <= NrOfMergingLanes; ++LaneNr) {
                      if (LaneNr == NrOfMergingLanes) {
                        g.setColor(fResources.kvisRampSeparationColor);
                      }
                      else {
                        g.setColor(fResources.kvisLaneSeparationColor);
                      }
                      fRenderer.plotLine(lri.fLanePositions[NodeNr][LaneNr - 1],
                                         lri.fRight[NodeNr + 1]);
                    }
                    g.setColor(fResources.kvisLaneSeparationColor);
                    // draw marks between corresponding lane-positions
                    for (int LaneNr = (NrOfMergingLanes + 1);
                         LaneNr <= lri.fLanePositions[NodeNr].length; ++LaneNr) {
                      fRenderer.plotLine(lri.fLanePositions[NodeNr][LaneNr - 1],
                                         lri.fLanePositions[NodeNr + 1][LaneNr - NrOfMergingLanes - 1]);
                    }
                  }
                }
                else {
                  // this section of the road is widened

                  if (link.fLocation == Link.kLeft) {

                    // draw marks between corresponding lane-positions
                    for (int LaneNr = 1; LaneNr <= lri.fLanePositions[NodeNr].length; ++LaneNr) {
                      fRenderer.plotLine(lri.fLanePositions[NodeNr][LaneNr - 1],
                                         lri.fLanePositions[NodeNr + 1][LaneNr - 1]);
                    }
                    // connect the last lane
                    g.setColor(fResources.kvisRampSeparationColor);
                    int LastLane = lri.fLanePositions[NodeNr].length + 1;
                    fRenderer.plotLine(lri.fLeft[NodeNr],
                                       lri.fLanePositions[NodeNr + 1][LastLane - 1]);
                    // draw the marks of the separating lanes
                    g.setColor(fResources.kvisLaneSeparationColor);
                    for (int LaneNr = (LastLane + 1); LaneNr <= lri.fLanePositions[NodeNr + 1].length; ++LaneNr) {
                      fRenderer.plotLine(lri.fLeft[NodeNr],
                                         lri.fLanePositions[NodeNr + 1][LaneNr - 1]);
                    }
                  }
                  else { // location is right
                    // draw the separating lanes
                    int NrOfSeparatingLanes = lri.fLanePositions[NodeNr + 1].length -
                                              lri.fLanePositions[NodeNr].length;
   
                    for (int LaneNr = 1; LaneNr <= (NrOfSeparatingLanes - 1); ++LaneNr) {
                      fRenderer.plotLine(lri.fRight[NodeNr],
                                         lri.fLanePositions[NodeNr + 1][LaneNr - 1]);
                    }
                    // connect the last lane
                    g.setColor(fResources.kvisRampSeparationColor);
                    int LastLane = NrOfSeparatingLanes;
                    fRenderer.plotLine(lri.fRight[NodeNr],
                                       lri.fLanePositions[NodeNr + 1][LastLane - 1]);
                    // draw marks between corresponding lane-positions
                    g.setColor(fResources.kvisLaneSeparationColor);
                    for (int LaneNr = 1; LaneNr <= lri.fLanePositions[NodeNr].length; ++LaneNr) {
                      fRenderer.plotLine(lri.fLanePositions[NodeNr][LaneNr - 1],
                                         lri.fLanePositions[NodeNr + 1][LaneNr + NrOfSeparatingLanes - 1]);
                    }
                  }
                } // a widened section
              } // section is narrowed or widened
            } // don't show congestion
          } // for (node)
        } // link-enum

        // show the vehicles in all the links
        if (fResources.fvisAnimateVehicles) {

          ViewLinkEnum = fViewLinkList.elements();
          while (ViewLinkEnum.hasMoreElements()) {
            Link link = (Link) ViewLinkEnum.nextElement();

            // show link in 3D
            LinkRenderingInformation lri = (LinkRenderingInformation)
                                           fLinkRenderingInformation.get(new Integer(link.fID));

            // draw the vehicles in this segment
            Enumeration VehicleEnum = link.fVehicleQueue.elements();
            while (VehicleEnum.hasMoreElements()) {
              Vehicle vehicle = (Vehicle) VehicleEnum.nextElement();

              int NodeNr = vehicle.fCurrentSegment;

              ViewportPoint vp1l = fRenderer.renderPoint(lri.fLeft[NodeNr]);
              ViewportPoint vp1r = fRenderer.renderPoint(lri.fRight[NodeNr]);
              ViewportPoint vp2l = fRenderer.renderPoint(lri.fLeft[NodeNr + 1]);
              ViewportPoint vp2r = fRenderer.renderPoint(lri.fRight[NodeNr + 1]);

              Point3D p1 = null;
              Point3D p2 = null;

              // is the vehicle is located on a straight segment (no narrowing or widening) ?
              if (lri.fLanePositions[NodeNr].length == lri.fLanePositions[NodeNr + 1].length) {

                if (vehicle.fCurrentLane == 0) {
                  p1 = lri.fRight[NodeNr];
                  p2 = lri.fRight[NodeNr + 1];
                }
                else {
                  p1 = lri.fLanePositions[NodeNr][vehicle.fCurrentLane - 1];
                  p2 = lri.fLanePositions[NodeNr + 1][vehicle.fCurrentLane - 1];
                }
              }
              else if (lri.fLanePositions[NodeNr].length > lri.fLanePositions[NodeNr + 1].length) {

                // the vehicle is located on a narrowing
                if (link.fLocation == Link.kLeft) {
                  if (vehicle.fCurrentLane == 0) {
                    p1 = lri.fRight[NodeNr];
                    p2 = lri.fRight[NodeNr + 1];
                  }
                  else {
                    p1 = lri.fLanePositions[NodeNr][vehicle.fCurrentLane - 1];
                    p2 = lri.fLanePositions[NodeNr + 1][vehicle.fCurrentLane - 1];
                  }
                }
                else { // the location of the onramp/narrowing is right
                  p1 = lri.fLanePositions[NodeNr][vehicle.fCurrentLane +
                                                  link.fNrOfLanesOnRamp - 1];
                  if (vehicle.fCurrentLane == 0) {
                    p2 = lri.fRight[NodeNr + 1];
                  }
                  else {
                    p2 = lri.fLanePositions[NodeNr + 1][vehicle.fCurrentLane - 1];
                  }
                }
              }
              else {
                // the vehicle is located on a widening
                if (link.fLocation == Link.kLeft) {
                  if (vehicle.fCurrentLane == 0) {
                    p1 = lri.fRight[NodeNr];
                    p2 = lri.fRight[NodeNr + 1];
                  }
                  else {
                    p1 = lri.fLanePositions[NodeNr][vehicle.fCurrentLane - 1];
                    p2 = lri.fLanePositions[NodeNr + 1][vehicle.fCurrentLane - 1];
                  }
                }
                else { // the location of the offramp/widening is right
                  if (vehicle.fCurrentLane == 0) {
                    p1 = lri.fRight[NodeNr];
                  }
                  else {
                    p1 = lri.fLanePositions[NodeNr][vehicle.fCurrentLane - 1];
                  }
                  p2 = lri.fLanePositions[NodeNr + 1][vehicle.fCurrentLane +
                                                      link.fNrOfLanesOnRamp - 1];
                }
              }

              // determine the z-coordinates of the vehicle's bottom
              // using a parametric equation
              float Ratio1 = (vehicle.fPositionOnCurrentSegment - vehicle.fLength) /
                             link.fLengths[NodeNr];
              float Ratio2 = vehicle.fPositionOnCurrentSegment /
                             link.fLengths[NodeNr];
              float z1 = p1.fZ + ((p2.fZ - p1.fZ) * Ratio1);
              float z2 = p1.fZ + ((p2.fZ - p1.fZ) * Ratio2);

              Point3D q = fRenderer.addPointByLine(p1,p2,vehicle.fPositionOnCurrentSegment,0.0f);

              Point3D p2r = fRenderer.addPointByLine(q,p1,
                                                     (fNetwork.fLaneWidth - vehicle.fWidth) / 2.0f,
                                                     (float) -(Math.PI / 2.0f));
              p2r.fZ = z2;

              Point3D p2l = fRenderer.addPointByLine(p2r,q,
                                                     vehicle.fWidth,
                                                     (float) Math.PI);
              p2l.fZ = z2;

              Point3D p1l = fRenderer.addPointByLine(p2l,p2r,
                                                     vehicle.fLength,
                                                     (float) -(Math.PI / 2.0f));
              p1l.fZ = z1;

              Point3D p1r = fRenderer.addPointByLine(p1l,p2l,
                                                     vehicle.fWidth,
                                                     (float) -(Math.PI / 2.0f));
              p1r.fZ = z1;

              Point3D q1l = new Point3D(p1l.fX,p1l.fY,p1l.fZ + vehicle.fHeight);
              Point3D q1r = new Point3D(p1r.fX,p1r.fY,p1r.fZ + vehicle.fHeight);
              Point3D q2l = new Point3D(p2l.fX,p2l.fY,p2l.fZ + vehicle.fHeight);
              Point3D q2r = new Point3D(p2r.fX,p2r.fY,p2r.fZ + vehicle.fHeight);

              if (fResources.fvisTrackingMode) {
                if (vehicle.fLicensePlate.equalsIgnoreCase(fResources.fvisLicensePlateToTrack)) {

                  fResources.fvisViewPointX = (int) ((p1l.fX + p1r.fX + p2l.fX + p2r.fX) / 4.0f);
                  fResources.fvisViewPointY = (int) ((p1l.fY + p1r.fY + p2l.fY + p2r.fY) / 4.0f);
                  fResources.fvisViewPointZ = (int) ((p1l.fZ + p2l.fZ + q1l.fZ + q2l.fZ) / 4.0f);
                }
              }

              ViewportPoint v1l = fRenderer.renderPoint(p1l);
              ViewportPoint v1r = fRenderer.renderPoint(p1r);
              ViewportPoint v2l = fRenderer.renderPoint(p2l);
              ViewportPoint v2r = fRenderer.renderPoint(p2r);

              ViewportPoint w1l = fRenderer.renderPoint(q1l);
              ViewportPoint w1r = fRenderer.renderPoint(q1r);
              ViewportPoint w2l = fRenderer.renderPoint(q2l);
              ViewportPoint w2r = fRenderer.renderPoint(q2r);

              if (fResources.fvisColorVehiclesByAggression) {
                g.setColor(fColorInterpolator.interpolate(
                           100 - ((int) Math.round(100.0f * vehicle.fDriver.fAggression))));
              }
              else {
                if (vehicle.fType == Vehicle.kCar) {
                  g.setColor(fResources.kvisCarColor);
                }
                else {
                  g.setColor(fResources.kvisTruckColor);
                }
              }

              // draw  the bottom-surface
              Polygon BottomSurface = new Polygon();
              BottomSurface.addPoint(v1l.fX,v1l.fY);
              BottomSurface.addPoint(v2l.fX,v2l.fY);
              BottomSurface.addPoint(v2r.fX,v2r.fY);
              BottomSurface.addPoint(v1r.fX,v1r.fY);
              g.fillPolygon(BottomSurface);

              // draw the top-surface
              Polygon TopSurface = new Polygon();
              TopSurface.addPoint(w1l.fX,w1l.fY);
              TopSurface.addPoint(w2l.fX,w2l.fY);
              TopSurface.addPoint(w2r.fX,w2r.fY);
              TopSurface.addPoint(w1r.fX,w1r.fY);
              g.fillPolygon(TopSurface);

              // draw the left-surface
              Polygon LeftSurface = new Polygon();
              LeftSurface.addPoint(v1l.fX,v1l.fY);
              LeftSurface.addPoint(w1l.fX,w1l.fY);
              LeftSurface.addPoint(w2l.fX,w2l.fY);
              LeftSurface.addPoint(v2l.fX,v2l.fY);
              g.fillPolygon(LeftSurface);

              // draw the right-surface
              Polygon RightSurface = new Polygon();
              RightSurface.addPoint(v1r.fX,v1r.fY);
              RightSurface.addPoint(w1r.fX,w1r.fY);
              RightSurface.addPoint(w2r.fX,w2r.fY);
              RightSurface.addPoint(v2r.fX,v2r.fY);
              g.fillPolygon(RightSurface);

              // draw the front-surface
              Polygon FrontSurface = new Polygon();
              FrontSurface.addPoint(v2l.fX,v2l.fY);
              FrontSurface.addPoint(w2l.fX,w2l.fY);
              FrontSurface.addPoint(w2r.fX,w2r.fY);
              FrontSurface.addPoint(v2r.fX,v2r.fY);
              g.fillPolygon(FrontSurface);

              // draw the back-surface
              Polygon BackSurface = new Polygon();
              BackSurface.addPoint(v1l.fX,v1l.fY);
              BackSurface.addPoint(w1l.fX,w1l.fY);
              BackSurface.addPoint(w1r.fX,w1r.fY);
              BackSurface.addPoint(v1r.fX,v1r.fY);
              g.fillPolygon(BackSurface);

              if (fResources.fvisShowMoreVehicleDetail) {

                if (vehicle.fType == Vehicle.kCar) {
                  g.setColor(fResources.kvisCarEdgeColor);
                }
                else {
                  g.setColor(fResources.kvisTruckEdgeColor);
                }
 
                // outline the vehicle's surfaces
                g.drawPolygon(BottomSurface);
                g.drawPolygon(TopSurface);
                g.drawPolygon(LeftSurface);
                g.drawPolygon(RightSurface);
                g.drawPolygon(FrontSurface);
                g.drawPolygon(BackSurface);

                // draw an arrow on the vehicle's roof indicating it's direction-of-travel
                Point3D u = fRenderer.addPointByLine(q1r,q2r,vehicle.fLength / 4.0f,0.0f);
                Point3D a1 = fRenderer.addPointByLine(u,q1r,vehicle.fWidth / 2.0f,
                                                      (float) -(Math.PI / 2.0f));
                Point3D a2 = fRenderer.addPointByLine(a1,u,2.0f * (vehicle.fLength / 4.0f),
                                                      (float) (Math.PI / 2.0f));
                Point3D a3 = fRenderer.addPointByLine(a2,a1,vehicle.fWidth,
                                                      (float) -(Math.PI / 8.0f));
                Point3D a4 = fRenderer.addPointByLine(a2,a1,vehicle.fWidth,
                                                      (float) (Math.PI / 8.0f));

                // determine the z-coordinates of the arrow using a parametric equation
                a1.fZ = q1r.fZ + ((q2r.fZ - q1r.fZ) * 0.25f);
                a2.fZ = q1r.fZ + ((q2r.fZ - q1r.fZ) * 0.75f);
                a3.fZ = q1r.fZ + ((q2r.fZ - q1r.fZ) * 0.5f);
                a4.fZ = q1r.fZ + ((q2r.fZ - q1r.fZ) * 0.5f);

                ViewportPoint va1 = fRenderer.renderPoint(a1);
                ViewportPoint va2 = fRenderer.renderPoint(a2);
                ViewportPoint va3 = fRenderer.renderPoint(a3);
                ViewportPoint va4 = fRenderer.renderPoint(a4);

                g.setColor(fResources.kvisDOTColor);
                g.drawLine(va1.fX,va1.fY,va2.fX,va2.fY);
                g.drawLine(va2.fX,va2.fY,va3.fX,va3.fY);
                g.drawLine(va2.fX,va2.fY,va4.fX,va4.fY);

                if (vehicle.fLeftBlinkerOn) {

                  // draw the left blinker
                  Point3D w = fRenderer.addPointByLine(p2l,p2r,vehicle.fWidth / 8.0f,0.0f);
                  Point3D lp2l = new Point3D(w.fX,w.fY,w.fZ + (3.0f * (vehicle.fHeight / 8.0f)));
                  Point3D lq2l = new Point3D(w.fX,w.fY,w.fZ + (5.0f * (vehicle.fHeight / 8.0f)));
                  w = fRenderer.addPointByLine(p2l,p2r,3.0f * (vehicle.fWidth / 8.0f),0.0f);
                  Point3D lp2r = new Point3D(w.fX,w.fY,w.fZ + (3.0f * (vehicle.fHeight / 8.0f)));
                  Point3D lq2r = new Point3D(w.fX,w.fY,w.fZ + (5.0f * (vehicle.fHeight / 8.0f)));

                  ViewportPoint vlp2l = fRenderer.renderPoint(lp2l);
                  ViewportPoint vlq2l = fRenderer.renderPoint(lq2l);
                  ViewportPoint vlp2r = fRenderer.renderPoint(lp2r);
                  ViewportPoint vlq2r = fRenderer.renderPoint(lq2r);

                  g.setColor(fResources.kvisBlinkerColor);

                  Polygon VehicleBlinker = new Polygon();
                  VehicleBlinker.addPoint(vlp2l.fX,vlp2l.fY);
                  VehicleBlinker.addPoint(vlq2l.fX,vlq2l.fY);
                  VehicleBlinker.addPoint(vlq2r.fX,vlq2r.fY);
                  VehicleBlinker.addPoint(vlp2r.fX,vlp2r.fY);
                  g.fillPolygon(VehicleBlinker);
                }

                if (vehicle.fRightBlinkerOn) {

                  // draw the right blinker
                  Point3D w = fRenderer.addPointByLine(p2l,p2r,5.0f * (vehicle.fWidth / 8.0f),0.0f);
                  Point3D lp2l = new Point3D(w.fX,w.fY,w.fZ + (3.0f * (vehicle.fHeight / 8.0f)));
                  Point3D lq2l = new Point3D(w.fX,w.fY,w.fZ + (5.0f * (vehicle.fHeight / 8.0f)));
                  w = fRenderer.addPointByLine(p2l,p2r,7.0f * (vehicle.fWidth / 8.0f),0.0f);
                  Point3D lp2r = new Point3D(w.fX,w.fY,w.fZ + (3.0f * (vehicle.fHeight / 8.0f)));
                  Point3D lq2r = new Point3D(w.fX,w.fY,w.fZ + (5.0f * (vehicle.fHeight / 8.0f)));

                  ViewportPoint vlp2l = fRenderer.renderPoint(lp2l);
                  ViewportPoint vlq2l = fRenderer.renderPoint(lq2l);
                  ViewportPoint vlp2r = fRenderer.renderPoint(lp2r);
                  ViewportPoint vlq2r = fRenderer.renderPoint(lq2r);

                  g.setColor(fResources.kvisBlinkerColor);

                  Polygon VehicleBlinker = new Polygon();
                  VehicleBlinker.addPoint(vlp2l.fX,vlp2l.fY);
                  VehicleBlinker.addPoint(vlq2l.fX,vlq2l.fY);
                  VehicleBlinker.addPoint(vlq2r.fX,vlq2r.fY);
                  VehicleBlinker.addPoint(vlp2r.fX,vlp2r.fY);
                  g.fillPolygon(VehicleBlinker);
                }

                // draw the stoplights
                if (vehicle.fCurrentAcceleration > 0.0f) {
                  g.setColor(Color.green);
                }
                else if (vehicle.fCurrentAcceleration < 0.0f) {
                  g.setColor(Color.red);
                }
                else {
                  if (fResources.fvisColorVehiclesByAggression) {
                    g.setColor(fColorInterpolator.interpolate(
                               100 - ((int) Math.round(100.0f * vehicle.fDriver.fAggression))));
                  }
                  else {
                    if (vehicle.fType == Vehicle.kCar) {
                      g.setColor(fResources.kvisCarColor);
                    }
                    else {
                      g.setColor(fResources.kvisTruckColor);
                    }
                  }
                }

                // draw the left stoplight
                Point3D w = fRenderer.addPointByLine(p1l,p1r,vehicle.fWidth / 8.0f,0.0f);
                Point3D lp1l = new Point3D(w.fX,w.fY,w.fZ + (3.0f * (vehicle.fHeight / 8.0f)));
                Point3D lq1l = new Point3D(w.fX,w.fY,w.fZ + (5.0f * (vehicle.fHeight / 8.0f)));
                w = fRenderer.addPointByLine(p1l,p1r,3.0f * (vehicle.fWidth / 8.0f),0.0f);
                Point3D lp1r = new Point3D(w.fX,w.fY,w.fZ + (3.0f * (vehicle.fHeight / 8.0f)));
                Point3D lq1r = new Point3D(w.fX,w.fY,w.fZ + (5.0f * (vehicle.fHeight / 8.0f)));

                ViewportPoint vlp1l = fRenderer.renderPoint(lp1l);
                ViewportPoint vlq1l = fRenderer.renderPoint(lq1l);
                ViewportPoint vlp1r = fRenderer.renderPoint(lp1r);
                ViewportPoint vlq1r = fRenderer.renderPoint(lq1r);

                Polygon VehicleLight = new Polygon();
                VehicleLight.addPoint(vlp1l.fX,vlp1l.fY);
                VehicleLight.addPoint(vlq1l.fX,vlq1l.fY);
                VehicleLight.addPoint(vlq1r.fX,vlq1r.fY);
                VehicleLight.addPoint(vlp1r.fX,vlp1r.fY);
                g.fillPolygon(VehicleLight);

                // draw the right stoplight
                w = fRenderer.addPointByLine(p1l,p1r,5.0f * (vehicle.fWidth / 8.0f),0.0f);
                lp1l = new Point3D(w.fX,w.fY,w.fZ + (3.0f * (vehicle.fHeight / 8.0f)));
                lq1l = new Point3D(w.fX,w.fY,w.fZ + (5.0f * (vehicle.fHeight / 8.0f)));
                w = fRenderer.addPointByLine(p1l,p1r,7.0f * (vehicle.fWidth / 8.0f),0.0f);
                lp1r = new Point3D(w.fX,w.fY,w.fZ + (3.0f * (vehicle.fHeight / 8.0f)));
                lq1r = new Point3D(w.fX,w.fY,w.fZ + (5.0f * (vehicle.fHeight / 8.0f)));

                vlp1l = fRenderer.renderPoint(lp1l);
                vlq1l = fRenderer.renderPoint(lq1l);
                vlp1r = fRenderer.renderPoint(lp1r);
                vlq1r = fRenderer.renderPoint(lq1r);

                VehicleLight = new Polygon();
                VehicleLight.addPoint(vlp1l.fX,vlp1l.fY);
                VehicleLight.addPoint(vlq1l.fX,vlq1l.fY);
                VehicleLight.addPoint(vlq1r.fX,vlq1r.fY);
                VehicleLight.addPoint(vlp1r.fX,vlp1r.fY);
                g.fillPolygon(VehicleLight);
              }

              // if necessary, show license plate, ... of the vehicle
              if (fResources.fvisShowVehicleInformation) {

                int MiddleX = (v1l.fX + v1r.fX + v2l.fX + v2r.fX) / 4;
                int MiddleY = (v1l.fY + v1r.fY + v2l.fY + v2r.fY) / 4;
                int xPos = ((v1r.fX + v2r.fX) / 2) + 35;
                int yPos = ((v1r.fY + v2r.fY) / 2) + 35;
                int TagWidth = 140;
                int TagHeight = 20;
                int Width = 160;
                int Height = 75;

                g.setColor(Color.darkGray);
                g.fillRect(xPos,yPos,TagWidth,TagHeight);
                g.fillRect(xPos,yPos + TagHeight,Width,TagHeight + Height);

                g.setColor(Color.white);
                g.drawRect(xPos,yPos,TagWidth,TagHeight);
                g.drawRect(xPos,yPos + TagHeight,Width,TagHeight + Height);
                g.drawLine(MiddleX,MiddleY,xPos + (TagWidth / 2),yPos);

                g.setColor(Color.cyan);
                g.drawString(vehicle.fLicensePlate,xPos + 5,yPos + 15);

                g.setColor(Color.white);
                g.drawString(String.valueOf(Math.round(vehicle.fDistanceTraveled)) +
                             " " + fMessages.lookup("textDistanceUnitAbbreviated"),
                             xPos + 5,yPos + TagHeight + 20);

                if (vehicle.fDesiredSpeedReached) {
                  g.setColor(Color.green);
                }
                else {
                  g.setColor(Color.red);
                }
                g.drawString(String.valueOf(Math.round(vehicle.fSpeed)) +
                             " " + fMessages.lookup("textSpeedUnitAbbreviated") + " / " +
                             String.valueOf(Math.round(vehicle.fDriver.fDesiredSpeed)) +
                             " " + fMessages.lookup("textSpeedUnitAbbreviated"),
                             xPos + 5,yPos + TagHeight + 35);

                g.setColor(Color.white);
                g.drawString(fMessages.lookup("textAggression") + " : " +
                             String.valueOf(Math.round(vehicle.fDriver.fAggression * 100.0f)) +
                             "%",
                             xPos + 5,yPos + TagHeight + 50);

                if (vehicle.fDriver.fMiddleLaneDriver) {
                  g.drawString(fMessages.lookup("textMiddleLaneDriver") + " : " +
                               fMessages.lookup("textYes"),
                               xPos + 5,yPos + TagHeight + 65);
                }
                else {
                  g.drawString(fMessages.lookup("textMiddleLaneDriver") + " : " +
                               fMessages.lookup("textNo"),
                               xPos + 5,yPos + TagHeight + 65);
                }

                if (vehicle.fMandatoryLocationReached) {
                  g.setColor(Color.green);
                }
                else {
                  g.setColor(Color.red);
                }
                g.fillRect(xPos + 5,yPos + TagHeight + 75,Width - 5 - 5,10);
              } // show information
            } // enumeration of vehicles in link
          } // enumeration of links
        } // don't show congestion
      }
      else { // show statistics

        // show the total number of vehicles in the simulator
        g.setColor(Color.white);
        g.drawString(fMessages.lookup("textNrOfVehiclesInSimulator") +
                     " : " + String.valueOf(fTotalNrOfVehicles),
                     ix1 + 10,iy1 + 60);

        g.setColor(Color.orange);
        g.drawLine(cx1 + 5,cy1 + 75,cx1 + 740,cy1 + 75);
        g.drawLine(cx1 + 5,cy1 + 95,cx1 + 740,cy1 + 95);

        g.setColor(Color.green);
        g.drawString(fMessages.lookup("textCaptionLoopDetector"),cx1 + 10,cy1 + 90);
        g.drawString(fMessages.lookup("textCaptionNrOfVehicles"),cx1 + 300,cy1 + 90);
        g.drawString(fMessages.lookup("textCaptionNrOfCars"),cx1 + 380,cy1 + 90);
        g.drawString(fMessages.lookup("textCaptionNrOfTrucks"),cx1 + 460,cy1 + 90);
        g.drawString(fMessages.lookup("textCaptionAverageSpeed"),cx1 + 545,cy1 + 90);
        g.drawString(fMessages.lookup("textCaptionAverageFlowRate"),cx1 + 605,cy1 + 90);
        g.drawString(fMessages.lookup("textCaptionAverageDensity"),cx1 + 685,cy1 + 90);

        // show the information gathered by the loopdetectors
        int yPosition = 125;
        Enumeration LoopDetectorEnum = fResources.fsimLoopDetectors.elements();
        while (LoopDetectorEnum.hasMoreElements()) {
          LoopDetector loopDetector = (LoopDetector) LoopDetectorEnum.nextElement();

          g.setColor(Color.yellow);
          g.drawString(loopDetector.fName,cx1 + 10,yPosition);

          g.setColor(Color.white);
          g.drawString(String.valueOf(loopDetector.fVehiclesPassed),cx1 + 300,yPosition);
          g.drawString(String.valueOf(loopDetector.fCarsPassed),cx1 + 380,yPosition);
          g.drawString(String.valueOf(loopDetector.fTrucksPassed),cx1 + 460,yPosition);
          g.drawString(String.valueOf(Math.round(loopDetector.fAverageSpeed)),cx1 + 540,yPosition);
          g.drawString(String.valueOf(Math.round(loopDetector.fAverageFlowRate)),cx1 + 590,yPosition);
          g.drawString(String.valueOf(Math.round(loopDetector.fAverageDensity)),cx1 + 670,yPosition);

          yPosition += 20;
        }

        g.setColor(Color.orange);
        g.drawLine(cx1 + 5,yPosition - 10,cx1 + 740,yPosition - 10);
        g.drawLine(cx1 + 5,cy1 + 75,cx1 + 5,yPosition - 10);
        g.drawLine(cx1 + 290,cy1 + 75,cx1 + 290,yPosition - 10);
        g.drawLine(cx1 + 370,cy1 + 75,cx1 + 370,yPosition - 10);
        g.drawLine(cx1 + 450,cy1 + 75,cx1 + 450,yPosition - 10);
        g.drawLine(cx1 + 530,cy1 + 75,cx1 + 530,yPosition - 10);
        g.drawLine(cx1 + 580,cy1 + 75,cx1 + 580,yPosition - 10);
        g.drawLine(cx1 + 660,cy1 + 75,cx1 + 660,yPosition - 10);
        g.drawLine(cx1 + 740,cy1 + 75,cx1 + 740,yPosition - 10);
      }

      // indicate the simulators' status
      g.setColor(Color.yellow);
      if (fSimulatorRunning) {
        g.drawString(fMessages.lookup("textSimulatorStatusRunning"),ix1 + 10,iy1 + 20);
      }
      else {
        g.drawString(fMessages.lookup("textSimulatorStatusStopped"),ix1 + 10,iy1 + 20);
      }

      // show the simulator's masterclock
      g.drawString(MasterClock.translate(fMasterClock),ix1 + 10,iy1 + 40);

      // draw the image on the canvas
      cg.drawImage(fCanvasImage,cx1 + 1,cy1 + 1,this);
    }
  }

  /** @author Sven Maerivoet<BR>
      All rights reserved.<BR>
      Universitaire Instelling Antwerpen
  */  
  // the message-canvas
  private class MessageCanvas extends Canvas
  {
    // the canvas-draw-area
    private int cx1;
    private int cy1;
    private int cx2;
    private int cy2;
    private int cwidth;
    private int cheight;

    public MessageCanvas(int width, int height)
    {
      super();

      setSize(width,height);
    }

    public void update(Graphics g)
    {
      paint(g);
    }

    public void paint(Graphics g)
    {
      // define the boundaries for drawing
      cx1 = 5;
      cy1 = 5;
      cx2 = (cx1 + getSize().width) - (getInsets().right + 5 + 5);
      cy2 = (cy1 + getSize().height) - (getInsets().bottom + 5 + 5);
      cwidth = cx2 - cx1;
      cheight = cy2 - cy1;

      // activate clipping of graphics primitives
      g.setClip(cx1,cy1,cwidth,cheight);
    }

    public void showMessage(String message)
    {
      Graphics g = getGraphics();
      g.clearRect(cx1,cy1,cwidth,cheight);
      g.setColor(fResources.knmMessageColor);
      g.drawString(message,cx1 + 5,cy1 + 18);
    }

    public void clearMessage()
    {
      showMessage("");
    }
  }

  /** @author Sven Maerivoet<BR>
      All rights reserved.<BR>
      Universitaire Instelling Antwerpen
  */
  private class LinkRenderingInformation
  {
    public Point3D[]   fLeft;
    public Point3D[]   fRight;
    public Point3D[][] fLanePositions;

    public void LinkRenderingInformation()
    {
      fLeft = null;
      fRight = null;
      fLanePositions = null;
    }
  }

  /** @author Sven Maerivoet<BR>
      All rights reserved.<BR>
      Universitaire Instelling Antwerpen
  */
  private class GeneralMenu extends Menu
  {
    public GeneralMenu(ActionListener listener, String menuName)
    {
      super(menuName);

      MenuItem mi;

      mi = new MenuItem(fMessages.lookup("menuAbout"));
      mi.addActionListener(listener);
      add(mi);

      mi = new MenuItem(fMessages.lookup("menuQuit"));
      mi.addActionListener(listener);
      add(mi);
    }
  } 

  /** @author Sven Maerivoet<BR>
      All rights reserved.<BR>
      Universitaire Instelling Antwerpen
  */
  private class DisplayMenu extends Menu implements ItemListener
  {
    // internal datastructures
    private CheckboxMenuItem fMenuAnimateVehicles;
    private CheckboxMenuItem fMenuShowStatistics;
    private CheckboxMenuItem fMenuShowTrafficFlow;
    private CheckboxMenuItem fMenuShowAverageSpeeds;
    private CheckboxMenuItem fMenuShowDensities;
    private CheckboxMenuItem fMenuShowVehicleInformation;
    private CheckboxMenuItem fMenuShowMoreVehicleDetail;
    private CheckboxMenuItem fMenuColorVehiclesByAggression;
    private CheckboxMenuItem fMenuShowLegend;
    private MenuItem         fMenuVehicleToTrack;

    public DisplayMenu(ActionListener listener, String menuName)
    {
      super(menuName);

      fMenuShowTrafficFlow = new CheckboxMenuItem(fMessages.lookup("menuShowTrafficFlow"));
      fMenuShowTrafficFlow.setState(fResources.fvisShowTrafficFlow);
      fMenuShowTrafficFlow.addItemListener(this);
      add(fMenuShowTrafficFlow);

      fMenuShowStatistics = new CheckboxMenuItem(fMessages.lookup("menuShowStatistics"));
      fMenuShowStatistics.setState(!fResources.fvisShowTrafficFlow);
      fMenuShowStatistics.addItemListener(this);
      add(fMenuShowStatistics);

      addSeparator();

      fMenuAnimateVehicles = new CheckboxMenuItem(fMessages.lookup("menuAnimateVehicles"));
      fMenuAnimateVehicles.setState(fResources.fvisAnimateVehicles);
      fMenuAnimateVehicles.addItemListener(this);
      add(fMenuAnimateVehicles);

      fMenuShowAverageSpeeds = new CheckboxMenuItem(fMessages.lookup("menuShowAverageSpeeds"));
      fMenuShowAverageSpeeds.setState(fResources.fvisShowAverageSpeeds);
      fMenuShowAverageSpeeds.addItemListener(this);
      add(fMenuShowAverageSpeeds);

      fMenuShowDensities = new CheckboxMenuItem(fMessages.lookup("menuShowDensities"));
      fMenuShowDensities.setState(fResources.fvisShowDensities);
      fMenuShowDensities.addItemListener(this);
      add(fMenuShowDensities);

      addSeparator();

      fMenuShowVehicleInformation = new CheckboxMenuItem(fMessages.lookup("menuShowVehicleInformation"));
      fMenuShowVehicleInformation.setState(fResources.fvisShowVehicleInformation);
      fMenuShowVehicleInformation.addItemListener(this);
      add(fMenuShowVehicleInformation);

      fMenuShowMoreVehicleDetail = new CheckboxMenuItem(fMessages.lookup("menuShowMoreVehicleDetail"));
      fMenuShowMoreVehicleDetail.setState(fResources.fvisShowMoreVehicleDetail);
      fMenuShowMoreVehicleDetail.addItemListener(this);
      add(fMenuShowMoreVehicleDetail);

      fMenuColorVehiclesByAggression = new CheckboxMenuItem(fMessages.lookup("menuColorVehiclesByAggression"));
      fMenuColorVehiclesByAggression.setState(fResources.fvisColorVehiclesByAggression);
      fMenuColorVehiclesByAggression.addItemListener(this);
      add(fMenuColorVehiclesByAggression);

      addSeparator();

      fMenuShowLegend = new CheckboxMenuItem(fMessages.lookup("menuShowLegend"));
      fMenuShowLegend.setState(fResources.fvisShowLegend);
      fMenuShowLegend.addItemListener(this);
      add(fMenuShowLegend);

      addSeparator();

      fMenuTrackingMode = new CheckboxMenuItem(fMessages.lookup("menuTrackingMode"));
      fMenuTrackingMode.setState(fResources.fvisTrackingMode);
      fMenuTrackingMode.addItemListener(this);
      add(fMenuTrackingMode);

      fMenuVehicleToTrack = new MenuItem(fMessages.lookup("menuVehicleToTrack"));
      fMenuVehicleToTrack.addActionListener(listener);
      add(fMenuVehicleToTrack);
    }

    // method from the ItemListener-interface
    public void itemStateChanged(ItemEvent e)
    {
      String Command = (e.getItem()).toString();

      if (Command.equals(fMessages.lookup("menuShowTrafficFlow"))) {
        fResources.fvisShowTrafficFlow = true;
      }
      if (Command.equals(fMessages.lookup("menuShowStatistics"))) {
        fResources.fvisShowTrafficFlow = false;
      }
      else if (Command.equals(fMessages.lookup("menuAnimateVehicles"))) {
        fResources.fvisAnimateVehicles = true;
        fResources.fvisShowAverageSpeeds = false;
        fResources.fvisShowDensities = false;
      }
      else if (Command.equals(fMessages.lookup("menuShowAverageSpeeds"))) {
        fResources.fvisAnimateVehicles = false;
        fResources.fvisShowAverageSpeeds = true;
        fResources.fvisShowDensities = false;
      }
      else if (Command.equals(fMessages.lookup("menuShowDensities"))) {
        fResources.fvisAnimateVehicles = false;
        fResources.fvisShowAverageSpeeds = false;
        fResources.fvisShowDensities = true;
      }
      else if (Command.equals(fMessages.lookup("menuShowVehicleInformation"))) {
        fResources.fvisShowVehicleInformation = !fResources.fvisShowVehicleInformation;
      }
      else if (Command.equals(fMessages.lookup("menuShowMoreVehicleDetail"))) {
        fResources.fvisShowMoreVehicleDetail = !fResources.fvisShowMoreVehicleDetail;
      }
      else if (Command.equals(fMessages.lookup("menuColorVehiclesByAggression"))) {
        fResources.fvisColorVehiclesByAggression = !fResources.fvisColorVehiclesByAggression;
      }
      else if (Command.equals(fMessages.lookup("menuShowLegend"))) {
        if (fLegend.isVisible()) {
          fLegend.setVisible(false);
        }
        else {
          fLegend.setVisible(true);
        }
      }
      else if (Command.equals(fMessages.lookup("menuTrackingMode"))) {
        fResources.fvisTrackingMode = !fResources.fvisTrackingMode;
      }

      if (fResources.fvisShowTrafficFlow) {

        fMenuAnimateVehicles.setEnabled(true);
        fMenuShowAverageSpeeds.setEnabled(true);
        fMenuShowDensities.setEnabled(true);
        fMenuShowVehicleInformation.setEnabled(true);
        fMenuShowMoreVehicleDetail.setEnabled(true);
        fMenuColorVehiclesByAggression.setEnabled(true);

        if (!fResources.fvisAnimateVehicles) {
          fMenuShowVehicleInformation.setEnabled(false);
          fMenuShowMoreVehicleDetail.setEnabled(false);
          fMenuColorVehiclesByAggression.setEnabled(false);
        }

        fMenuShowTrafficFlow.setState(true);
        fMenuShowStatistics.setState(false);
      }
      else { // show statistics

        fMenuAnimateVehicles.setEnabled(false);
        fMenuShowAverageSpeeds.setEnabled(false);
        fMenuShowDensities.setEnabled(false);
        fMenuShowVehicleInformation.setEnabled(false);
        fMenuShowMoreVehicleDetail.setEnabled(false);
        fMenuColorVehiclesByAggression.setEnabled(false);

        fMenuShowTrafficFlow.setState(false);
        fMenuShowStatistics.setState(true);
      }

      if (fResources.fvisAnimateVehicles) {
        fMenuAnimateVehicles.setState(true);
      }
      else {
        fMenuAnimateVehicles.setState(false);
      }

      if (fResources.fvisShowAverageSpeeds) {
        fMenuShowAverageSpeeds.setState(true);
      }
      else {
        fMenuShowAverageSpeeds.setState(false);
      }

      if (fResources.fvisShowDensities) {
        fMenuShowDensities.setState(true);
      }
      else {
        fMenuShowDensities.setState(false);
      }

      if (fResources.fvisShowVehicleInformation) {
        fMenuShowVehicleInformation.setState(true);
      }
      else {
        fMenuShowVehicleInformation.setState(false);
      }

      if (fResources.fvisShowMoreVehicleDetail) {
        fMenuShowMoreVehicleDetail.setState(true);
      }
      else {
        fMenuShowMoreVehicleDetail.setState(false);
      }

      if (fResources.fvisColorVehiclesByAggression) {
        fMenuColorVehiclesByAggression.setState(true);
      }
      else {
        fMenuColorVehiclesByAggression.setState(false);
      }
    }
  } 

  /** @author Sven Maerivoet<BR>
      All rights reserved.<BR>
      Universitaire Instelling Antwerpen
  */
  private class LoggingMenu extends Menu
  {
    public LoggingMenu(ActionListener listener, String menuName)
    {
      super(menuName);

      fStartLogging = new MenuItem(fMessages.lookup("menuStartLogging"));
      fStartLogging.addActionListener(listener);
      add(fStartLogging);
      fStartLogging.setEnabled(false);

      fStopLogging = new MenuItem(fMessages.lookup("menuStopLogging"));
      fStopLogging.addActionListener(listener);
      add(fStopLogging);
      fStopLogging.setEnabled(false);
    }
  } 
}
